from gsdriver.base.abstract import CustomDict
from gsdriver.base.types import _KT, _VT, Status, Datetime, Data

from gsdriver.utils.cast import cast_numeric, cast_str
from gsdriver.utils.date import get_datetime, get_datetime_pair, get_date, get_date_pair, get_weekday
from gsdriver.utils.map import notna_dict, filter_data, regex_get, safe_len
from gsdriver.utils.map import aggregate_data as agg

from typing import Any, Dict, List, Literal, Optional, Sequence, Tuple, Union
import datetime as dt
import os
import re
import sys

import jwt
import requests
import time


DateFormat = Union[str,Tuple[str,str]]

HEADER = "header"
FOOTER = "footer"
NULL = "#N/A"

ENV_PATH = "env/"
PRIVATE_KEY = ENV_PATH+"private.key"

NAVER_WORKS_LIMIT = 2000
NATEON_LIMIT = 2000

DEFAULT_DATE_FORMAT = "%y-%m-%d({d})"
DEFAULT_DATE_END_FORMAT = "%m-%d-({d})"

DEFAULT_DATETIME_FORMAT = "%y-%m-%d({d}) %H:%M:%S"
DEFAULT_DATETIME_END_FORMAT = "%H:%M:%S"

DEFAULT_RANGE_FORMAT = "{start} to {end}"

DEFAULT_TEXT_BODY_FORMAT = "▶ [{name}] {length():,} rows collected"


###################################################################
########################### Format Text ###########################
###################################################################

class Mapping(dict):
    def __missing__(self, key: str) -> Any:
        return str() if key in (HEADER,FOOTER) else NULL


def format_map(__s: str, mapping: Dict) -> str:
    return __s.format_map(Mapping(mapping))


def get_format_map(data: Data, textFormat: str, textHeader: Optional[str]=None, textFooter: Optional[str]=None,
                dateFormat: Optional[DateFormat]=None, rangeFormat: Optional[str]=None,
                tzinfo=None, busdate=False, **kwargs) -> Dict:
    mapping = dict()
    if ("{header}" in textFormat) and isinstance(textHeader, str):
        textFormat = textFormat.replace("{header}", textHeader)
    if ("{footer}" in textFormat) and isinstance(textFooter, str):
        textFormat = textFormat.replace("{footer}", textFooter)
    for __key in set(regex_get(r"\{([^}]+)\}", textFormat, indices=[])):
        try:
            __key, __value = format_key_value(data, __key, dateFormat, rangeFormat, tzinfo, busdate, **kwargs)
            mapping[__key] = __value
        except: pass
    return mapping


def format_key_value(data: Data, __key: str, dateFormat: Optional[DateFormat]=None, rangeFormat: Optional[str]=None,
                    tzinfo=None, busdate=False, **kwargs) -> Tuple[_KT,_VT]:
    if re.match(r"^(date|datetime)\(.+\)$", __key):
        dateFunc, path = regex_get(r"^(\w+)\((.+)\)$", __key, groups=[0,1])
        __datetime = _get_date_object(kwargs.get(path), dateFunc, tzinfo=tzinfo, busdate=busdate)
        return __key, format_date(__datetime, (dateFormat if dateFormat else dateFunc), tzinfo=tzinfo)
    elif re.match(r"^(daterange|timerange)\([^,]+,[^,]+\)$", __key):
        dateFunc, start, end = regex_get(r"^(\w+)\(([^,]+),([^,]+)\)$", __key, groups=[0,1,2])
        __start, __end = _get_date_range_object(kwargs.get(start), kwargs.get(end), dateFunc, tzinfo=tzinfo, busdate=busdate)
        return __key, format_date_range(__start, __end, (dateFormat if dateFormat else dateFunc), rangeFormat, tzinfo=tzinfo)
    else: return __key.rsplit(':', maxsplit=1)[0], format_value(data, __key, **kwargs)


def format_value(data: Data, __key: str, **kwargs) -> _VT:
    __key, isNumber = (__key.rsplit(':', maxsplit=1)[0], True) if ':' in __key else (__key, False)
    if re.match(r"^\w+\(.*\)$", __key):
        aggFunc, path = regex_get(r"^(\w+)\((.*)\)$", __key, groups=[0,1])
        return aggregate_data(data, aggFunc, path)
    elif __key in kwargs:
        return cast_numeric(kwargs[__key]) if isNumber else cast_str(kwargs[__key], default=NULL)
    else: return 0 if isNumber else NULL


def aggregate_data(data: Data, aggFunc: str, path: str) -> Union[float,int]:
    if '>' in path:
        name, column = path.split('>', maxsplit=1)
        return aggregate_data(data[name], aggFunc, column)
    elif aggFunc == "length":
        return safe_len(data[path]) if path else safe_len(data)
    else:
        values = filter_data(data, path, if_null="drop") if path else data
        return cast_numeric(agg(values, aggFunc))


def split_lines(__s: str, maxLength: int, sep='\n', strip=True) -> List[str]:
    array, lines, length = list(), str(), 0
    for line in __s.split(sep):
        length += len(line)+len(sep)
        if length >= maxLength:
            array.append(lines.strip() if strip else lines)
            lines, length = line, len(line)
        else: lines = sep.join([lines, line]) if lines else line
    if lines: array.append(lines.strip() if strip else lines)
    return array


###################################################################
########################### Format Date ###########################
###################################################################

def format_date(__datetime: Datetime, dateFormat: str, tzinfo=None) -> str:
    dateFormat = get_date_format(dateFormat)
    if not (isinstance(__datetime, (dt.date,dt.datetime)) and dateFormat):
        return NULL
    elif ("{d}" in dateFormat) or ("{ddd}" in dateFormat):
        mapping = {__k: get_weekday(__datetime, __f, tzinfo) for __k, __f in zip(["d","ddd"],["short","long"])}
        dateFormat = format_map(dateFormat, mapping)
    return __datetime.strftime(dateFormat.encode("unicode-escape").decode()).encode().decode("unicode-escape")


def format_date_range(__start: Datetime, __end: Datetime, dateFormat: DateFormat, rangeFormat=str(), tzinfo=None) -> str:
    (startFormat, endFormat), rangeFormat = get_date_range_format(dateFormat, rangeFormat)
    if __start == __end:
        return format_date(__start, dateFormat=startFormat, tzinfo=tzinfo)
    elif not (isinstance(__start, (dt.date,dt.datetime)) and (type(__start) == type(__end)) and startFormat and endFormat):
        return format_map(rangeFormat, dict(start=NULL, end=NULL))
    if isinstance(__start, dt.datetime) and re.search(r"(%[^\s]*H|%[^\s]*M|%[^\s]*S)", startFormat):
        endFormat = endFormat if __start.date() == __end.date() else startFormat
    else: endFormat = endFormat if __start.year == __end.year else startFormat
    __start = format_date(__start, dateFormat=startFormat, tzinfo=tzinfo)
    __end = format_date(__end, dateFormat=endFormat, tzinfo=tzinfo)
    return rangeFormat.format_map(dict(start=__start, end=__end))


def get_date_format(dateFormat=str()) -> str:
    if dateFormat == "date": return DEFAULT_DATE_FORMAT
    elif dateFormat == "datetime": return DEFAULT_DATETIME_FORMAT
    else: return dateFormat if isinstance(dateFormat, str) else str()


def get_date_end_format(dateFormat=str()) -> str:
    if dateFormat in "date": return DEFAULT_DATE_END_FORMAT
    elif dateFormat == "datetime": return DEFAULT_DATETIME_END_FORMAT
    else: return dateFormat if isinstance(dateFormat, str) else str()


def get_date_range_format(dateFormat: Optional[DateFormat]=None, rangeFormat=str()) -> Tuple[Tuple[str,str],str]:
    rangeFormat = rangeFormat if rangeFormat else DEFAULT_RANGE_FORMAT
    if dateFormat == "daterange": return (DEFAULT_DATE_FORMAT, DEFAULT_DATE_END_FORMAT), rangeFormat
    elif dateFormat == "timerange": return (DEFAULT_DATETIME_FORMAT, DEFAULT_DATETIME_END_FORMAT), rangeFormat
    elif isinstance(dateFormat, str):
        return (get_date_format(dateFormat), get_date_end_format(dateFormat)), rangeFormat
    elif isinstance(dateFormat, Sequence) and (len(dateFormat) == 2):
        return (get_date_format(dateFormat[0]), get_date_end_format(dateFormat[1])), rangeFormat
    else: return (str(), str()), rangeFormat


def _get_date_object(__object, dateFunc: Literal["date","datetime"]=str(), tzinfo=None, busdate=False) -> Datetime:
    if isinstance(__object, (dt.date,dt.datetime)): return __object
    elif dateFunc == "date": return get_date(__object, if_null=None, tzinfo=tzinfo, busdate=busdate)
    elif dateFunc == "datetime": return get_datetime(__object, if_null=None, tzinfo=tzinfo)
    else: return None


def _get_date_range_object(__start, __end, dateFunc: Literal["daterange","timerange"]=str(),
                            tzinfo=None, busdate=False) -> Tuple[Datetime,Datetime]:
    if dateFunc == "daterange": return get_date_pair(__start, __end, if_null=None, tzinfo=tzinfo, busdate=busdate)
    elif dateFunc == "timerange": return get_datetime_pair(__start, __end, if_null=None, tzinfo=tzinfo)
    else: return (None, None)


###################################################################
########################### Private Key ###########################
###################################################################

def get_sys_file(file: str) -> str:
    base_path = getattr(sys, '_MEIPASS', os.path.abspath(os.path.dirname(__file__)))
    return os.path.join(base_path, file)


def read_file(file: str) -> Union[str,bytes]:
    if os.path.exists(file):
        with open(file, "rb") as f:
            return f.read()
    else: return file


def read_private_key(file=str()) -> Union[str,bytes]:
    if isinstance(file, str) and file.endswith(".key") and os.path.exists(file): pass
    elif os.path.exists(PRIVATE_KEY): file = PRIVATE_KEY
    else: file = get_sys_file(PRIVATE_KEY.split('/')[-1])
    return read_file(file)


###################################################################
######################### NAVER WORKS API #########################
###################################################################

class AlertInfo(CustomDict):
    def __init__(self, clientId: str, clientSecret: str, serviceAccount: str, botId: str,
                channelId=str(), userId=str(), limit=2000, strip=True, privateKey: Union[str,bytes]=str(),
                textFormat: Optional[str]=None, textHeader: Optional[str]=None, textBody: Optional[str]=None, textFooter: Optional[str]=None,
                dateFormat: Optional[DateFormat]=None, rangeFormat: Optional[str]=None, name: Optional[str]=None):
        authInfo = dict(clientId=clientId, clientSecret=clientSecret, serviceAccount=serviceAccount, privateKey=privateKey)
        botInfo = dict(botId=botId, channelId=channelId, userId=userId, limit=min(limit,NAVER_WORKS_LIMIT), strip=strip)
        textInfo = notna_dict(name=name, dateFormat=dateFormat, rangeFormat=rangeFormat)
        textInfo.update(self.set_text_format(textFormat, textHeader, textBody, textFooter))
        super().__init__(authInfo=authInfo, botInfo=botInfo, textInfo=textInfo)

    def set_text_format(self, textFormat=None, textHeader=None, textBody=None, textFooter=None) -> Dict[str,str]:
        textFormat = textFormat if textFormat else ("{header}"+(textBody if textBody else DEFAULT_TEXT_BODY_FORMAT)+"{footer}")
        return notna_dict(textFormat=textFormat, textHeader=textHeader, textFooter=textFooter)


class NaverBot(requests.Session):
    accessToken = str()

    def __init__(self, clientId: str, clientSecret: str, serviceAccount: str, privateKey: Union[str,bytes]=str()):
        super().__init__()
        privateKey = privateKey if isinstance(privateKey, bytes) else read_private_key(privateKey)
        self.accessToken = self.authorize(clientId, clientSecret, serviceAccount, privateKey)

    def authorize(self, clientId: str, clientSecret: str, serviceAccount: str, privateKey: Union[str,bytes], expires=60*5) -> str:
        url = "https://auth.worksmobile.com/oauth2/v2.0/token"
        params = dict(
            assertion=self.encode_jwt(clientId, serviceAccount, privateKey, expires),
            grant_type="urn:ietf:params:oauth:grant-type:jwt-bearer",
            client_id=clientId,
            client_secret=clientSecret,
            scope="bot"
        )
        response = self.post(url, params=params)
        return response.json()["access_token"]

    def encode_jwt(self, clientId: str, serviceAccount: str, privateKey: Union[str,bytes], expires=60*5) -> str:
        headers = {"alg":"RS256", "typ":"JWT"}
        payload = {"iss":clientId, "sub":serviceAccount, "iat":int(time.time()), "exp":int(time.time())+expires}
        return jwt.encode(payload, privateKey, algorithm="HS256", headers=headers)

    def send_text(self, text: str, botId: str, channelId=str(), userId=str(), limit=2000, strip=True) -> Status:
        __id, __type = (channelId if channelId else userId), ("channels" if channelId else "users")
        url = f"https://www.worksapis.com/v1.0/bots/{botId}/{__type}/{__id}/messages"
        headers = {"Authorization":f"Bearer {self.accessToken}", "Content-Type":"application/json"}
        status = list()
        for lines in split_lines((text.strip() if strip else text), maxLength=limit, strip=strip):
            data = {"content":{"type":"text", "text":lines}}
            response = self.post(url, headers=headers, json=data)
            status.append(response.status_code)
        return status[0] if len(status) == 1 else status


###################################################################
############################ NateOn API ###########################
###################################################################

def alert_nateon(nateonUrl: str, content: str, limit=(2000-100)) -> Status:
    status, headers = list(), {"Content-Type":"application/x-www-form-urlencoded"}
    for lines in split_lines(content, maxLength=limit):
        data = {"content":lines}
        status.append(requests.post(nateonUrl, headers=headers, data=data).status_code)
    return status[0] if len(status) == 1 else status



from gsdriver.base.types import LogLevel, TabularData, Data

from typing import Any, Dict, List, Optional, Sequence, Union
from aiohttp import ClientResponse
from requests import Response
import logging

from bs4 import Tag
from pandas import DataFrame, Series
import json
import re
import sys
import traceback


LOG_FILE = "log"
CRITICAL = 50
FATAL = CRITICAL
ERROR = 40
WARNING = 30
WARN = WARNING
INFO = 20
DEBUG = 10
NOTSET = 0

LOG_FORMAT = "%(asctime)s [%(levelname)s] <%(name)s> %(module)s:%(lineno)d %(funcName)s() | %(message)s"
LOG_FORMAT2 = "[%(levelname)s] <%(name)s> %(module)s:%(lineno)d %(funcName)s() | %(message)s"

JSON_LOG_FORMAT = '{"asctime": "%(asctime)s", "level": "%(levelname)s", "name": "%(name)s", \
"module": "%(module)s", "line": %(lineno)d, "func": "%(funcName)s", "message": %(message)s},'


def get_default_log_format(file=str()) -> str:
    if not (isinstance(file, str) and file): return LOG_FORMAT2
    elif file.endswith(".json"): return JSON_LOG_FORMAT
    else: return LOG_FORMAT


###################################################################
############################## Logger #############################
###################################################################

class CustomLogger(logging.Logger):
    def __init__(self, name=__name__, level: LogLevel=WARN, file=str(), fmt=str(), datefmt="%Y-%m-%d %H:%M:%S"):
        super().__init__(name, level)
        formatter = logging.Formatter(fmt=(fmt if fmt else get_default_log_format(file)), datefmt=datefmt)
        handler = logging.FileHandler(file, encoding="utf-8") if file else logging.StreamHandler()
        handler.setLevel(level)
        handler.setFormatter(formatter)
        self.addHandler(handler)


def get_log_level(level: Union[int,str]) -> int:
    if isinstance(level, str):
        if not level: return None
        elif level.isdigit(): return int(level)
        else: return attr if isinstance(attr := getattr(logging, level.upper(), None), int) else None
    else: return int(level) if isinstance(level, (float,int)) else None


def format_json_log(path: str):
    with open(path, "r", encoding="utf-8") as file:
        logs = file.read().rstrip()
        logs = re.sub(r",*\n]\n", ",\n", logs)
        if logs[0] != '[':
            logs = '[\n' + logs
        if logs[-1] != ']':
            logs = (logs[:-1] if logs[-1] == ',' else logs) + '\n]\n'
    with open(path, "w", encoding="utf-8") as file:
        file.write(logs)


###################################################################
########################### Log Message ###########################
###################################################################

def _limit(__object: Union[Data,str], limit=1000) -> Union[Data,str]:
    if (not isinstance(limit, int)) or (len(str(__object)) < limit):
        return __object
    elif isinstance(__object, str):
        return __object[:limit] + "..."
    elif isinstance(__object, List):
        if len(str(__object[:5])) < limit: return __object[:5] + ["..."]
        else: return [_limit(__object[0], limit), "..."]
    elif isinstance(__object, Dict):
        if len(str(__object)) < limit: return __object
        __object = {__key: "..." for __key in __object.keys()}
        return __object if len(str(__object)) < limit else {"...": "..."}
    else: return str(__object)[:limit]


def limit_data(__object, limit=1000, depth=2, data_only=True) -> Union[Any,Dict]:
    if isinstance(__object, str):
        return _limit(__object, limit) if data_only else {"data-type":"str", "data-length":len(__object), "data":_limit(__object, limit)}
    elif isinstance(__object, Dict):
        __data = _limit((__object if depth < 1 else
            {__key: limit_data(__value, limit=limit, depth=(depth-1)) for __key, __value in __object.items()}), limit)
        return __data if data_only else {"data-type":"Dict", "dict-keys":len(__object), "data":__data}
    elif isinstance(__object, Sequence):
        __data = _limit((__object if depth < 1 else
            [(limit_data(__e, limit=limit, depth=(depth-1)) if isinstance(__e, Dict) else __e) for __e in __object]), limit)
        __length = [len(__object), len(__data[0])] if __data and isinstance(__data[0], Dict) else len(__object)
        return __data if data_only else {"data-type":"List", "data-length":__length, "data":__data}
    elif isinstance(__object, DataFrame):
        if depth > 0: __data = limit_data(__object.to_dict("records"), limit=limit, depth=(depth-1))
        else: __data = _limit(__object.to_dict("records"), limit)
        return __data if data_only else {"data-type":"DataFrame", "data-shape":list(__object.shape), "data":__data}
    elif isinstance(__object, Series):
        if depth > 0: __data = limit_data(__object.tolist(), limit=limit, depth=(depth-1))
        else: __data = _limit(__object.tolist(), limit)
        return __data if data_only else {"data-type":"Series", "data-length":len(__object), "data":__data}
    elif isinstance(__object, Tag):
        __object = __object.prettify()
        return _limit(__object, limit) if data_only else {"data-type":"Tag", "data-length":len(__object), "data":_limit(__object, limit)}
    else: return _limit(__object, limit) if data_only else {"data-type":str(type(__object)), "data":_limit(__object, limit)}


def log_object(object: Dict, indent=None, ensure_ascii=False, default=str, **options) -> str:
    return json.dumps(object, indent=indent, ensure_ascii=ensure_ascii, default=default, **options)


def log_encrypt(show=3, **kwargs) -> str:
    encrypt = lambda string, show=3: str(string)[:show].ljust(len(str(string)),'*')
    return log_object(dict(**{key:encrypt(value, show=show) for key, value in kwargs.items()}))


def log_messages(params: Optional[Dict]=None, data: Optional[Dict]=None, json: Optional[Dict]=None,
                headers: Optional[Dict]=None, cookies=None, **kwargs) -> str:
    params = dict(params=limit_data(params, limit=3000)) if params else dict()
    data = dict(data=limit_data((data if data else json), limit=3000)) if data or json else dict()
    headers = dict(headers=limit_data(headers, limit=3000)) if headers else dict()
    cookies = dict(cookies=limit_data(cookies, limit=3000)) if cookies else dict()
    return log_object(dict(**limit_data(kwargs), **data, **params, **headers, **cookies))


def log_response(response: Response, url: str, **kwargs) -> str:
    try: length = len(response.text)
    except: length = None
    return log_object(dict(**kwargs, **{"url":url, "status":response.status_code, "contents-length":length}))


async def log_client(response: ClientResponse, url: str, **kwargs) -> str:
    try: length = len(await response.text())
    except: length = None
    return log_object(dict(**kwargs, **{"url":url, "status":response.status, "contents-length":length}))


def log_data(data: Data, **kwargs) -> str:
    return log_object(dict(**limit_data(kwargs), data=limit_data(data, data_only=False)))


def log_error(func: str, **kwargs) -> str:
    error = ''.join(traceback.format_exception(*sys.exc_info()))
    return log_object(dict(func=func, kwargs=limit_data(kwargs), error=error))


def log_table(data: TabularData, schema: Optional[List]=None, **kwargs) -> str:
    schema = dict(schema=limit_data(schema, limit=3000)) if schema else dict()
    if not isinstance(data, (DataFrame,Series)):
        try: shape = (len(data),)
        except: shape = (0,)
    else: shape = data.shape
    return log_object(dict(**kwargs, **{"table-shape":shape}, **schema))



from __future__ import annotations
from gsdriver.base.abstract import OptionalDict, TypedRecords, Value, ValueSet
from gsdriver.base.abstract import GCLOUD_CONTEXT, INVALID_OBJECT_MSG, INVALID_MSG, INVALID_OBJECT_TYPE_MSG
from gsdriver.base.session import BaseSession

from gsdriver.base.types import _KT, Context, TypeHint, IndexLabel, DateFormat, RenameMap, TypeMap
from gsdriver.base.types import TabularData, PostData, from_literal

from gsdriver.utils.cast import cast_float, cast_list, cast_date, cast_datetime, cast_datetime_format
from gsdriver.utils.date import get_datetime, DATE_UNIT
from gsdriver.utils.logs import log_table
from gsdriver.utils.map import isna, df_empty, to_array, kloc, to_dict, to_records, read_table
from gsdriver.utils.map import cloc, to_dataframe, convert_data, rename_data, filter_data, apply_data

from google.oauth2 import service_account
from google.oauth2.service_account import IDTokenCredentials
from google.auth.transport.requests import AuthorizedSession

from gspread.worksheet import Worksheet
import gspread

from google.cloud import bigquery
from google.cloud.bigquery.table import RowIterator
from google.cloud.bigquery.job import LoadJobConfig

from tqdm import tqdm
import abc
import copy
import functools
import os
import requests
import sys

from typing import Any, Dict, Iterable, List, Literal, Optional, Sequence, Tuple, Union
import datetime as dt
import json
import re
import pandas as pd


Account = Union[Dict[str,str], str]

ENV_PATH = "env/"
GCLOUD_ACCOUNT = ENV_PATH+"gcloud.json"
GCLOUD_DATA = ENV_PATH+"data.json"

NAME = "name"
KEY, SHEET, FIELDS = "key", "sheet", "fields"
TABLE, QUERY, PID = "table", "query", "project_id"
MODE, DATA = "mode", "data"

FILEPATH, SHEETNAME = "file_path", "sheet_name"

FROM_GS, TO_GBQ = ["from_key", "from_sheet"], ["to_table", "to_pid"]
FROM_GBQ, TO_GS = ["from_query", "from_pid"], ["to_key", "to_sheet"]

READ = lambda name=str(): f"read_{name}" if name else "read"
UPLOAD = lambda name=str(): f"upload_{name}" if name else "upload"


###################################################################
############################# Messages ############################
###################################################################

UPLOAD_FAILED_MSG = "There is no upload data."
UPLOAD_GBQ_MSG = lambda table: f"Uploading data to '{table}' table"

UPLOAD_STATUS_TITLE = "▷ Upload Status"
UPLOAD_STATUS_MSG = lambda idx, name, type, status=True: f'({idx}) "{name}" {type} {"✅" if status else "❌"}'
GOOGL_SHEETS = "sheets"
BIGQUERY_TABLE = "table"

GOOGLE_READ_CONTEXT = "Google cloud read context"
GOOGLE_QUERY_CONTEXT = "Google cloud query context"
GOOGLE_UPLOAD_CONTEXT = "Google cloud upload context"

BIGQUERY_TYPE = "BigQueryType"
BIGQUERY_MODE = "BigQueryMode"
BIGQUERY_SCHEMA = "BigQuerySchema"

PARTITION_FILTER = "Partition filter"


###################################################################
###################### Google Authorization #######################
###################################################################

def get_sys_file(file: str) -> str:
    base_path = getattr(sys, '_MEIPASS', os.path.abspath(os.path.dirname(__file__)))
    return os.path.join(base_path, file)


def read_json(file: str) -> Dict:
    if os.path.exists(file):
        with open(file, "r", encoding="utf-8") as f:
            return json.loads(f.read())
    else: return dict()


def read_gcloud(file=str()) -> Dict:
    if isinstance(file, str) and file.endswith(".json") and os.path.exists(file): pass
    elif os.path.exists(GCLOUD_ACCOUNT): file = GCLOUD_ACCOUNT
    else: file = get_sys_file(GCLOUD_ACCOUNT.split('/')[-1])
    return read_json(file)


def read_data(operation: str, file=str()) -> PostData:
    return read_json(str(file) if str(file).endswith(".json") else GCLOUD_DATA).get(operation, dict())


def fetch_gcloud_credentials(audience=str(), account: Account=dict()) -> IDTokenCredentials:
    account = account if account and isinstance(account, dict) else read_gcloud(account)
    audience = audience if audience else account.get("audience", str())
    return service_account.IDTokenCredentials.from_service_account_info(account, target_audience=audience)


def fetch_gcloud_authorization(audience=str(), account: Account=dict()) -> str:
    credentials = fetch_gcloud_credentials(audience, account)
    auth_session = AuthorizedSession(credentials)
    auth_session.get(audience)
    return "Bearer "+credentials.token


def request_gcloud(audience: str, data: Optional[PostData]=dict(), authorization=str(),
                    account: Account=dict(), file=str(), operation=str()) -> requests.Response:
    if not authorization:
        authorization = fetch_gcloud_authorization(audience, account)
    data = data if data and isinstance(data, dict) else read_data(file).get(operation, dict())
    return requests.post(audience, json=data, headers={"Authorization":authorization})


###################################################################
######################### BigQuery Schema #########################
###################################################################

BigQueryNumericType = Literal["INTEGER", "FLOAT", "NUMERIC", "BIGNUMERIC", "BOOLEAN"]
BigQUeryDatetimeType = Literal["TIMESTAMP", "DATE", "TIME", "DATETIME"]
BigQueryDataType = Literal["GEOGRAPHY", "RECORD", "JSON"]
BigQueryType = Union[Literal["STRING", "BYTES"], BigQueryNumericType, BigQUeryDatetimeType, BigQueryDataType]

BigQueryMode = Literal["NULLABLE", "REQUIRED", "REPEATED"]

class BigQueryField(Value):
    typeCast = False

    def __init__(self, name: str, type: BigQueryType, mode: Optional[BigQueryMode]=None,
                description: Optional[str]=None, maxLength: Optional[int]=None):
        super().__init__(name=name, **self.validate_type(type),
            optional=dict(**self.validate_mode(mode), description=description, maxLength=maxLength))

    def validate_type(self, type: BigQueryType) -> Context:
        if type not in from_literal(BigQueryType):
            raise ValueError(INVALID_OBJECT_MSG(type, BIGQUERY_TYPE))
        else: return dict(type=type)

    def validate_mode(self, mode: BigQueryMode) -> Context:
        if mode and (mode not in from_literal(BigQueryMode)):
            raise ValueError(INVALID_OBJECT_MSG(mode, BIGQUERY_MODE))
        else: return dict(mode=mode)

    def copy(self) -> BigQueryField:
        return copy.deepcopy(self)

    def update(self, __m: Dict=dict(), inplace=True, self_var=False, **kwargs) -> BigQueryField:
        return super().update(__m, inplace=inplace, self_var=self_var, **kwargs)


class BigQuerySchema(ValueSet):
    dtype = BigQueryField
    typeCheck = True

    def __init__(self, *fields: BigQueryField):
        super().__init__(*fields)

    def copy(self) -> BigQuerySchema:
        return copy.deepcopy(self)

    def update(self, __iterable: Iterable[BigQueryField], inplace=True) -> BigQuerySchema:
        return super().update(__iterable, inplace=inplace)

    def map(self, key: str, value: str) -> Dict:
        key, value = re.sub(r"^desc$", "description", key), re.sub(r"^desc$", "description", value)
        return super().map(key, value)

    def get_primary_key(self) -> IndexLabel:
        return [field.get("name") for field in self if isinstance(field, Dict) and field.get("mode") == "REQUIRED"]


def validate_gbq_schema(schema: Any, optional=False) -> BigQuerySchema:
    if optional and (not schema): return
    elif isinstance(schema, BigQuerySchema): return schema
    elif isinstance(schema, List): return BigQuerySchema(*schema)
    else: raise TypeError(INVALID_OBJECT_TYPE_MSG(schema, BIGQUERY_SCHEMA))


###################################################################
######################## Google Cloud Query #######################
###################################################################

NumericiseIgnore = Union[Sequence[int], bool]

class GspreadReadContext(OptionalDict):
    def __init__(self, key: str, sheet: str, fields: Optional[IndexLabel]=None, default: Optional[Any]=None,
                if_null: Literal["drop","pass"]="pass", head=1, headers: Optional[IndexLabel]=None,
                str_cols: Optional[NumericiseIgnore]=None, return_type: Optional[TypeHint]="dataframe",
                rename: Optional[RenameMap]=None, to: Optional[Literal["desc","name"]]="name",
                convert_dtypes=True, size: Optional[int]=None, name=str(), **kwargs):
        super().__init__(key=key, sheet=sheet,
            optional=dict(
                fields=fields, default=default, if_null=if_null, head=head, headers=headers,
                str_cols=str_cols, return_type=return_type, rename=rename, to=to, convert_dtypes=convert_dtypes,
                size=size, name=name, **kwargs),
            null_if=dict(if_null="pass", head=1, to="name", return_type="dataframe", convert_dtypes=True, name=str()))


class GspreadQueryContext(GspreadReadContext):
    def __init__(self, key: str, sheet: str, fields: Optional[IndexLabel]=None, default: Optional[Any]=None,
                if_null: Literal["drop","pass"]="drop", axis=0, dropna=True, drop_empty=False, unique=False,
                head=1, headers: Optional[IndexLabel]=None, str_cols: Optional[NumericiseIgnore]=None,
                arr_cols: Optional[IndexLabel]=None, as_records=False, as_frame=False,
                rename: Optional[RenameMap]=None, to: Optional[Literal["desc","name"]]="name",
                convert_dtypes=True, size: Optional[int]=None, name=str(), **kwargs):
        return_type = "records" if as_records else "dataframe"
        super().__init__(
            key, sheet, fields, default, if_null, head, headers, str_cols, return_type, rename, to, convert_dtypes, size, name)
        self.update_notna(
                axis=axis, dropna=dropna, drop_empty=drop_empty, unique=unique, arr_cols=arr_cols,
                as_records=as_records, as_frame=as_frame, **kwargs,
            null_if=dict(axis=0, dropna=True, drop_empty=False, unique=False, as_records=False, as_frame=False))


class BigQueryReadContext(OptionalDict):
    def __init__(self, query: str, project_id: str, return_type: Optional[TypeHint]="dataframe",
                size: Optional[int]=None, name=str(), **kwargs):
        super().__init__(query=query, project_id=project_id,
            optional=dict(return_type=return_type, size=size, name=name, **kwargs),
            null_if=dict(return_type="dataframe", name=str()))


class BigQueryContext(BigQueryReadContext):
    def __init__(self, query: str, project_id: str, axis=0, dropna=True, drop_empty=False, unique=False,
                arr_cols: Optional[IndexLabel]=None, as_records=False, as_frame=False,
                size: Optional[int]=None, name=str(), **kwargs):
        return_type = "records" if as_records else "dataframe"
        super().__init__(query, project_id, return_type, size, name)
        self.update_notna(
                axis=axis, dropna=dropna, drop_empty=drop_empty, unique=unique, arr_cols=arr_cols,
                as_records=as_records, as_frame=as_frame, **kwargs,
            null_if=dict(axis=0, dropna=True, drop_empty=False, unique=False, as_records=False, as_frame=False))


class ExcelReadContext(OptionalDict):
    def __init__(self, file_path: str, sheet_name: Union[str,int]=0, fields: Optional[IndexLabel]=None,
                default: Optional[Any]=None, if_null: Literal["drop","pass"]="pass",
                dtype: Optional[Union[TypeMap,Literal["str"]]]=None, return_type: Optional[TypeHint]="dataframe",
                rename: Optional[RenameMap]=None, to: Optional[Literal["desc","name"]]="name",
                regex_path=False, reverse_path=False, size: Optional[int]=None, name=str(), **kwargs):
        super().__init__(file_path=file_path, sheet_name=sheet_name,
            optional=dict(
                fields=fields, default=default, if_null=if_null, dtype=(str if dtype == "str" else dtype), return_type=return_type,
                rename=rename, to=to, regex_path=regex_path, reverse_path=reverse_path, size=size, name=name, **kwargs),
            null_if=dict(if_null="pass", return_type="dataframe", to="name", regex_path=False, reverse_path=False, name=str()))


class ExcelQueryContext(ExcelReadContext):
    def __init__(self, file_path: str, sheet_name: Union[str,int]=0, fields: Optional[IndexLabel]=None,
                default: Optional[Any]=None, if_null: Literal["drop","pass"]="pass", axis=0,
                dropna=True, drop_empty=False, unique=False, dtype: Optional[Union[TypeMap,Literal["str"]]]=None,
                arr_cols: Optional[IndexLabel]=None, as_records=False, as_frame=False,
                rename: Optional[RenameMap]=None, to: Optional[Literal["desc","name"]]="name",
                regex_path=False, reverse_path=False, size: Optional[int]=None, name=str(), **kwargs):
        return_type = "records" if as_records else "dataframe"
        super().__init__(
            file_path, sheet_name, fields, default, if_null, dtype, return_type, rename, to, regex_path, reverse_path, size, name)
        self.update_notna(
                axis=axis, dropna=dropna, drop_empty=drop_empty, unique=unique, arr_cols=arr_cols,
                as_records=as_records, as_frame=as_frame, **kwargs,
            null_if=dict(axis=0, dropna=True, drop_empty=False, unique=False, as_records=False, as_frame=False))


class GoogleReadContext(GspreadReadContext, BigQueryReadContext, ExcelReadContext):
    def __init__(self, **kwargs):
        if len(kloc(kwargs, [KEY, SHEET], if_null="drop")) == 2: self.__class__ = GspreadReadContext
        elif len(kloc(kwargs, [QUERY, PID], if_null="drop")) == 2: self.__class__ = BigQueryReadContext
        elif FILEPATH in kwargs: self.__class__ = ExcelReadContext
        else: raise ValueError(INVALID_MSG(GOOGLE_READ_CONTEXT))
        self.__class__.__init__(self, **kwargs)


class GoogleQueryContext(GspreadQueryContext, BigQueryContext, ExcelQueryContext):
    def __init__(self, **kwargs):
        if len(kloc(kwargs, [KEY, SHEET], if_null="drop")) == 2: self.__class__ = GspreadQueryContext
        elif len(kloc(kwargs, [QUERY, PID], if_null="drop")) == 2: self.__class__ = BigQueryContext
        elif FILEPATH in kwargs: self.__class__ = ExcelQueryContext
        else: raise ValueError(INVALID_MSG(GOOGLE_QUERY_CONTEXT))
        self.__class__.__init__(self, **kwargs)


class GoogleQueryList(TypedRecords):
    dtype = GoogleQueryContext
    typeCheck = True

    def __init__(self, *args: GoogleQueryContext):
        super().__init__(*args)

    def validate_dtype(self, __object) -> Dict:
        if isinstance(__object, (GspreadQueryContext, BigQueryContext, ExcelQueryContext)): return __object
        elif isinstance(__object, Dict): return GoogleQueryContext(**__object)
        else: self.raise_dtype_error(__object)


class GoogleQueryReader(BaseSession):
    __metaclass__ = abc.ABCMeta
    operation = "googleQueryReader"

    def read_data(self, context: Dict, account: Account=dict()) -> TabularData:
        if isinstance(context, GspreadReadContext): return self.read_gspread(account=account, **context)
        elif isinstance(context, BigQueryReadContext): return self.read_gbq(account=account, **context)
        elif isinstance(context, ExcelReadContext): return self.read_excel(account=account, **context)
        else: return self.read_data(GoogleReadContext(**context), account)

    def read_gspread(self, key: str, sheet: str, fields: IndexLabel=list(), default=None,
                    if_null: Literal["drop","pass"]="pass", head=1, headers=None,
                    str_cols: NumericiseIgnore=list(), return_type: Optional[TypeHint]="dataframe",
                    rename: Optional[RenameMap]=None, to: Optional[Literal["desc","name"]]="name",
                    convert_dtypes=True, size: Optional[int]=None, name=str(), account: Account=dict(), **context) -> TabularData:
        rename_map = rename if rename else self.get_rename_map(to=to, query=True)
        kwargs = dict(numericise_ignore=str_cols, return_type=return_type, rename=rename_map, convert_dtypes=convert_dtypes, account=account)
        data = read_gspread(key, sheet, fields, default, if_null, head, headers, **kwargs)
        if isinstance(size, int): data = data[:size]
        self.checkpoint(READ(name), where="read_gspread", msg={KEY:key, SHEET:sheet, DATA:data}, save=data)
        self.logger.info(log_table(data, name=name, key=key, sheet=sheet))
        return data

    def read_gbq(self, query: str, project_id: str, return_type: Literal["records","dataframe"]="dataframe",
                size: Optional[int]=None, name=str(), account: Account=dict(), **context) -> TabularData:
        data = read_gbq(query, project_id, return_type=return_type, account=account)
        if isinstance(size, int): data = data[:size]
        self.checkpoint(READ(name), where="read_gspread", msg={QUERY:query, PID:project_id, DATA:data}, save=data)
        self.logger.info(log_table(data, name=name, query=query, project_id=project_id))
        return data

    def read_excel(self, file_path: str, sheet_name: Union[str,int]=0,
                    fields: IndexLabel=list(), default=None, if_null: Literal["drop","pass"]="pass",
                    dtype: TypeMap=None, return_type: Literal["records","dataframe"]="dataframe",
                    rename: Optional[RenameMap]=None, to: Optional[Literal["desc","name"]]="name",
                    regex_path=False, reverse_path=False, size: Optional[int]=None, name=str(), **context) -> TabularData:
        rename_map = rename if rename else self.get_rename_map(to=to, query=True)
        kwargs = dict(dtype=dtype, return_type=return_type, rename=rename_map, regex_path=regex_path, reverse_path=reverse_path)
        data = read_table(file_path, sheet_name=sheet_name, columns=fields, default=default, if_null=if_null, **kwargs)
        if isinstance(size, int): data = data[:size]
        self.checkpoint(READ(name), where="read_excel", msg={FILEPATH:file_path, SHEETNAME:sheet_name, DATA:data}, save=data)
        self.logger.info(log_table(data, name=name, file_path=file_path, sheet_name=sheet_name))
        return data

    def set_query(self, queryList: GoogleQueryList=list(), account: Account=dict()):
        for queryContext in GoogleQueryList(*queryList):
            data = self.read_data(queryContext, account)
            self.update(self.map_query_data(data, **queryContext), inplace=True)

    def map_query_data(self, data: pd.DataFrame, axis=0, dropna=True, drop_empty=False, unique=False,
                        arr_cols: IndexLabel=list(), as_records=False, as_frame=False,
                        name=str(), **context) -> Dict[_KT,Union[List,Any]]:
        if as_records or as_frame: return {name: data}
        arr_cols = cast_list(arr_cols)
        __m = to_dict((data.T if axis == 1 else data), "list", depth=2)
        for __key, __values in __m.copy().items():
            values = to_array(__values, dropna=dropna, drop_empty=drop_empty, unique=unique)
            if (len(values) < 2) and (__key not in arr_cols):
                __m[__key] = values[0] if len(values) == 1 else None
            else: __m[__key] = values
        return __m


###################################################################
###################### Google Cloud Uploader ######################
###################################################################

class GspreadUpdateContext(OptionalDict):
    def __init__(self, key: str, sheet: str, columns: IndexLabel=list(),
                mode: Literal["append","replace","ignore","upsert"]="append", primary_key: Optional[_KT]=None,
                set_date: Optional[Dict]=None, cell: Optional[str]=None, read: Optional[Dict]=None, name=str(), **kwargs):
        super().__init__(key=key, sheet=sheet,
            optional=dict(
                columns=columns, mode=mode, primary_key=cast_list(primary_key), set_date=set_date,
                cell=cell, read=read, name=name, **kwargs),
            null_if=dict(mode="append", head=1, to="name", name=str()))


class BigQueryPartition(OptionalDict):
    def __init__(self, field: str, type: Literal["date","datetime","number"]="date",
                unit: Literal["auto","hour","day","month","year"]="auto",
                left=None, right=None, value=None, limited=True, drop_partition=False, filter_partition=True):
        has_value = (left is not None) or (right is not None) or (value is not None)
        super().__init__(field=field, type=type, unit=unit.lower(),
            optional=dict(
                left=left, right=right, value=value, limited=limited,
                drop_partition=drop_partition, filter_partition=filter_partition, has_value=has_value))


class BigQueryUploadContext(OptionalDict):
    def __init__(self, table: str, project_id: str, columns: IndexLabel=list(),
                mode: Literal["append","replace","ignore","upsert"]="append", primary_key: Optional[_KT]=None,
                partition: Optional[Dict]=None, set_date: Optional[Dict]=None, read: Optional[Dict]=None, name=str(), **kwargs):
        super().__init__(table=table, project_id=project_id,
            optional=dict(
                columns=columns, mode=mode, primary_key=cast_list(primary_key),
                partition=partition, set_date=set_date, read=read, name=name, **kwargs),
            null_if=dict(mode="append", partition_by="auto", name=str()))


class GoogleUploadContext(GspreadUpdateContext, BigQueryUploadContext):
    def __init__(self, **kwargs):
        if len(kloc(kwargs, [KEY, SHEET], if_null="drop")) == 2: self.__class__ = GspreadUpdateContext
        elif len(kloc(kwargs, [TABLE, PID], if_null="drop")) == 2: self.__class__ = BigQueryUploadContext
        else: raise ValueError(INVALID_MSG(GOOGLE_UPLOAD_CONTEXT))
        self.__class__.__init__(self, **kwargs)


class GoogleCopyContext(OptionalDict):
    def __init__(self, read: Dict, upload: Dict, name=str(), **kwargs):
        if not isinstance(read, (GspreadReadContext, BigQueryReadContext, ExcelReadContext)):
            read = GoogleReadContext(**(read if isinstance(read, Dict) else dict()))
            if "name" not in read: read["name"] = name
        if not isinstance(upload, (GspreadUpdateContext, BigQueryUploadContext)):
            upload = GoogleUploadContext(**(upload if isinstance(upload, Dict) else dict()))
            if "name" not in upload: upload["name"] = name
        super().__init__(read=read, upload=upload)


class GoogleUploadList(TypedRecords):
    dtype = GoogleUploadContext
    typeCheck = True

    def __init__(self, *args: GoogleUploadContext):
        super().__init__(*args)

    def validate_dtype(self, __object) -> Dict:
        if isinstance(__object, (GspreadUpdateContext, BigQueryUploadContext, GoogleCopyContext)): return __object
        elif isinstance(__object, Dict):
            if ("read" in __object) and ("upload" in __object): return GoogleCopyContext(**__object)
            else: return GoogleUploadContext(**__object)
        else: self.raise_dtype_error(__object)


class GoogleUploader(GoogleQueryReader):
    __metaclass__ = abc.ABCMeta
    operation = "googleUploader"
    uploadStatus = dict()

    def upload_data(self, data: TabularData, uploadList: GoogleUploadList=list(), account: Account=dict(), **context):
        data = to_dataframe(data)
        context = GCLOUD_CONTEXT(**context)
        for uploadContext in GoogleUploadList(*uploadList):
            target = self._get_upload_target(uploadContext)
            if (not data.empty) and isinstance(uploadContext, GspreadUpdateContext):
                self.uploadStatus[target] = self.upload_gspread(data=data.copy(), **uploadContext, account=account, **context)
            elif (not data.empty) and isinstance(uploadContext, BigQueryUploadContext):
                self.uploadStatus[target] = self.upload_gbq(data=data.copy(), **uploadContext, account=account, **context)
            elif isinstance(uploadContext, GoogleCopyContext):
                self.uploadStatus[target] = self.copy_data(**uploadContext, account=account, **context)
            else: self.uploadStatus[target] = False

    def _get_upload_target(self, uploadContext) -> str:
        if isinstance(uploadContext, GspreadUpdateContext): return uploadContext["sheet"]
        elif isinstance(uploadContext, BigQueryUploadContext): return uploadContext["table"]
        elif isinstance(uploadContext, GoogleCopyContext): return self._get_upload_target(uploadContext["upload"])
        else: return uploadContext.get(NAME, "unnamed") if isinstance(uploadContext, Dict) else "unnamed"

    @BaseSession.ignore_exception
    def copy_data(self, read: Dict, upload: Dict, account: Account=dict(), **context) -> bool:
        read["rename"] = read.get("rename", self.get_rename_map(to=context.get("to", "name")))
        data = self.read_data(read, account)
        self.upload_data(data, [upload], account, **context)
        return True

    def format_upload_status(self, **context) -> str:
        formatted = list()
        for __idx, (__name, __status) in enumerate(self.uploadStatus.items(), start=1):
            __type = BIGQUERY_TABLE if re.match(r"^[\w\d]+\.[\w\d]+$", __name) else GOOGL_SHEETS
            formatted.append(UPLOAD_STATUS_MSG(__idx, __name, __type, __status))
        return '\n'.join([UPLOAD_STATUS_TITLE]+formatted) if formatted else str()

    ###################################################################
    ###################### Google Spread Sheets #######################
    ###################################################################

    @BaseSession.ignore_exception
    def upload_gspread(self, key: str, sheet: str, data: pd.DataFrame, columns: IndexLabel=list(),
                        mode: Literal["append","replace","ignore","upsert"]="append", primary_key: List[str]=list(),
                        set_date=None, cell=str(), read=None, name=str(), account: Account=dict(), **context) -> bool:
        params = dict(key=key, sheet=sheet, mode=mode, primary_key=primary_key, set_date=set_date, name=name, account=account)
        if mode in ("ignore","upsert"):
            read = self.set_base_sheet(read=read, **params, **context)
            mode = {"ignore":"append", "upsert":"replace"}.get(mode)
        data = self.from_base_data(data, read, **params, **context)
        data = self.map_upload_data(data, columns=columns, **params, **context)
        self.checkpoint(UPLOAD(name), where="upload_gspread", msg={KEY:key, SHEET:sheet, MODE:mode, DATA:data}, save=data)
        self.logger.info(log_table(data, name=name, key=key, sheet=sheet, mode=mode))
        cell = "A2" if mode == "replace" else (cell if cell else str())
        update_gspread(key, sheet, data, cell=cell, clear=(mode == "replace"), account=account)
        return True

    def set_base_sheet(self, key: str, sheet: str, read=None, name=str(), **context) -> GoogleReadContext:
        if isinstance(read, (BigQueryReadContext, ExcelReadContext)): return read
        read = read if isinstance(read, Dict) else dict()
        for __key, __value in zip([KEY, SHEET, NAME], [key, sheet, name]):
            if __key not in read: read[__key] = __value
        return read if isinstance(read, GspreadReadContext) else GspreadReadContext(**read)

    ###################################################################
    ######################### Google BigQuery #########################
    ###################################################################

    @BaseSession.ignore_exception
    def upload_gbq(self, table: str, project_id: str, data: pd.DataFrame, columns: IndexLabel=list(),
                    mode: Literal["append","replace","ignore","upsert"]="append", primary_key: List[str]=list(),
                    partition=None, set_date=None, read=None, progress=True, name=str(), account: Account=dict(), **context) -> bool:
        partition = self.validate_partition(partition, **context)
        params = dict(table=table, project_id=project_id, mode=mode,
            primary_key=primary_key, partition=partition, set_date=set_date, name=name, account=account)
        if mode in ("ignore","upsert"):
            read = self.set_base_query(read=read, **params, **context)
            mode = {"ignore":"append", "upsert":"replace"}.get(mode)
        data = self.from_base_data(data, read, **params, **context)
        data = self.map_upload_data(data, columns=columns, **params, **context)
        self.checkpoint(UPLOAD(name), where="upload_gbq", msg={TABLE:table, PID:project_id, MODE:mode, DATA:data}, save=data)
        self.logger.info(log_table(data, name=name, table=table, pid=project_id, mode=mode))
        upload_gbq(table, project_id, data, mode, partition, progress, account)
        return True

    def validate_partition(self, partition=None, startDate=None, endDate=None, busdate=False, **context) -> BigQueryPartition:
        if not (isinstance(partition, Dict) and partition.get("field")):
            return dict()
        elif (partition.get("type") in ("date","datetime")) and (partition.get("unit") != "hour"):
            if ("value" not in partition):
                left, right = self.get_date_pair(partition.get("left", startDate), partition.get("right", endDate), busdate=busdate)
                partition["left"], partition["right"] = left, right
            else: partition["value"] = self.get_date(partition["value"])
        return partition if isinstance(partition, BigQueryPartition) else BigQueryPartition(**partition)

    def set_base_query(self, table: str, project_id: str, read=None, partition=dict(), name=str(), **context) -> GoogleReadContext:
        if isinstance(read, (GspreadReadContext, ExcelReadContext)): return read
        read = read if isinstance(read, Dict) else dict()
        read = self.set_partition_query(read, table, partition=partition, name=name, **context)
        for __key, __value in zip([PID, NAME], [project_id, name]):
            if __key not in read: read[__key] = __value
        return read if isinstance(read, BigQueryReadContext) else BigQueryReadContext(**read)

    def set_partition_query(self, read: Dict, table: str, partition=dict(), **context) -> Dict:
        if QUERY in read: return read
        else: read[QUERY] = make_select_query(table, **partition)
        return read

    ###################################################################
    ########################## From Base Data #########################
    ###################################################################

    def from_base_data(self, data: pd.DataFrame, read: Optional[Dict]=None, mode=str(),
                        primary_key: List[str]=list(), name=str(), account: Account=dict(), **context) -> pd.DataFrame:
        if not isinstance(read, Dict): return data
        base = to_dataframe(self.read_data(read, account))
        if mode == "ignore":
            return self.map_ignore_data(data, base, primary_key=primary_key, name=name, **context)
        elif mode == "upsert":
            return self.map_upsert_data(data, base, primary_key=primary_key, name=name, **context)
        else: return self.map_upload_base(data, base, primary_key=primary_key, name=name, **context)

    def map_upload_base(self, data: pd.DataFrame, base: pd.DataFrame, name=str(), **context) -> pd.DataFrame:
        return data

    def map_ignore_data(self, data: pd.DataFrame, base: pd.DataFrame, primary_key: List[str], name=str(), **context) -> pd.DataFrame:
        if (not primary_key) or base.empty: return data
        data = data.merge(base[primary_key].drop_duplicates(primary_key), how="left", on=primary_key, indicator=True)
        return data[data["_merge"] == "left_only"].drop(columns=["_merge"])

    def map_upsert_data(self, data: pd.DataFrame, base: pd.DataFrame, primary_key: List[str],
                        agg_func: Dict=dict(), name=str(), **context) -> pd.DataFrame:
        if (not primary_key) or base.empty: return data
        agg_func = agg_func if agg_func else {__column:"first" for __column in base.columns if __column not in primary_key}
        return pd.concat([data, base]).groupby(primary_key).agg(agg_func).reset_index()

    ###################################################################
    ######################### Map Upload Data #########################
    ###################################################################

    def map_upload_data(self, data: pd.DataFrame, columns: IndexLabel=list(), primary_key: List[str]=list(),
                        set_date=None, name=str(), **context) -> pd.DataFrame:
        columns = cast_list(columns) if columns else self.get_upload_columns(name=name, **context)
        if columns:
            data = cloc(data, columns, if_null="pass", reorder=True)
        if primary_key:
            data = data.dropna(subset=primary_key, how="any").drop_duplicates(primary_key)
        if isinstance(set_date, Dict) and set_date:
            data = self.set_upload_date(data, set_date=set_date, name=name, **context)
        if data.empty:
            raise ValueError(UPLOAD_FAILED_MSG)
        return data

    def get_upload_columns(self, name=str(), **context) -> IndexLabel:
        return list()

    def set_upload_date(self, data: pd.DataFrame, set_date: Dict[str,DateFormat], name=str(), **context) -> pd.DataFrame:
        data = data.copy()
        for __key, __value in set_date.items():
            __date = self.get_date(__value)
            if (__key in data) and isinstance(__date, dt.date):
                data[__key] = __date
        return data


###################################################################
###################### Google Spread Sheets #######################
###################################################################

def load_gspread(key: str, sheet: str, account: Account=dict()) -> Worksheet:
    account = account if account and isinstance(account, dict) else read_gcloud(account)
    gs_acc = gspread.service_account_from_dict(account)
    gs = gs_acc.open_by_key(key)
    return gs.worksheet(sheet)


def gs_loaded(func):
    @functools.wraps(func)
    def wrapper(key=str(), sheet=str(), *args, account: Account=dict(), gs: Optional[Worksheet]=None, **kwargs):
        if not gs:
            gs = load_gspread(key, sheet, account)
        return func(key, sheet, *args, account=account, gs=gs, **kwargs)
    return wrapper


@gs_loaded
def read_gspread(key: str, sheet: str, fields: Optional[IndexLabel]=list(), default=None,
                if_null: Literal["drop","pass"]="pass", head=1, headers=None, numericise_ignore: NumericiseIgnore=list(),
                reorder=True, return_type: Optional[TypeHint]="dataframe", rename: RenameMap=dict(), convert_dtypes=True,
                account: Account=dict(), gs: Optional[Worksheet]=None) -> TabularData:
    if isinstance(numericise_ignore, bool):
        numericise_ignore = ["all"] if numericise_ignore else list()
    data = gs.get_all_records(head=head, default_blank=default, numericise_ignore=numericise_ignore, expected_headers=headers)
    data = convert_data(data, return_type)
    data = rename_data(data, rename)
    data = filter_data(data, cast_list(fields), default=default, if_null=if_null, reorder=reorder)
    return apply_data(data, apply=_convert_dtype, all_keys=True) if convert_dtypes else data


@gs_loaded
def clear_gspead(key: str, sheet: str, include_header=False, account: Account=dict(), gs: Optional[Worksheet]=None):
    if include_header: return gs.clear()
    last_row = len(gs.get_all_records())+1
    gs.insert_row([], 2)
    gs.delete_rows(3, last_row+2)


@gs_loaded
def update_gspread(key: str, sheet: str, data: TabularData, col='A', row=0, cell=str(), clear=False,
                    clear_header=False, account: Account=dict(), gs: Optional[Worksheet]=None):
    records = to_records(data)
    if not records: return
    values = [[_to_excel_date(__value) for __value in __m.values()] for __m in records]
    if clear:
        clear_gspead(gs=gs, include_header=clear_header)
    if not cell:
        cell = col+str(row if row else len(gs.get_all_records())+2)
        gs.add_rows(len(values))
    gs.update(cell, values)


def _convert_dtype(__value) -> Any:
    if isinstance(__value, str):
        if __value == "TRUE": return True
        elif __value == "FALSE": return False
        elif re.match(r"^\d[\d.]*%$", __value): return cast_float(__value) / 100
        else: return cast_datetime_format(__value, default=__value)
    else: return __value


def _to_excel_date(__value) -> Union[int,Any]:
    if isna(__value): return None
    elif not isinstance(__value, dt.date): return __value
    offset = 693594
    days = __value.toordinal() - offset
    if isinstance(__value, dt.datetime):
        seconds = (__value.hour*60*60 + __value.minute*60 + __value.second)/(24*60*60)
        return days + seconds
    return days


###################################################################
######################### Google BigQuery #########################
###################################################################

BIGQUERY_JOB = {"append":"WRITE_APPEND", "replace":"WRITE_TRUNCATE"}

class PartitionedIterator:
    data = None
    field = str()
    index = 0
    partition = list()
    drop_partition = False

    def __init__(self, data: pd.DataFrame, field: str, type: Literal["date","datetime","number"]="date",
                unit: Literal["auto","hour","day","month","year"]="auto", left=None, right=None, value=None,
                drop_partition=False, filter_partition=True, has_value=False, **kwargs):
        data, isCreated = self.create_partition(data, field, type, unit)
        self.data = data
        self.field = "_PARTITIONTIME" if isCreated else field
        self.index = 0
        self.partition = self.set_partition(data, self.field, type, left, right, value, (filter_partition and has_value))
        self.drop_partition = True if isCreated else drop_partition

    def create_partition(self, data: pd.DataFrame, field: str, type: Literal["date","datetime","number"]="date",
                        unit: Literal["auto","hour","day","month","year"]="auto") -> Tuple[pd.DataFrame,bool]:
        if type in ("date","datetime"):
            data = data.copy()
            if (type == "datetime") or (unit != "auto"):
                data["_PARTITIONTIME"] = data[field].apply(lambda x: get_datetime(x, unit=(unit if unit in DATE_UNIT else "day")))
                return data[data["_PARTITIONTIME"].notna()], True
            else: data[field] = data[field].apply(cast_date)
        return data[data[field].notna()], False

    def set_partition(self, data: pd.DataFrame, field: str, type: Literal["date","datetime","number"]="date",
                    left=None, right=None, value=None, filter_partition=True) -> List:
        if filter_partition:
            if field == "_PARTITIONTIME": left, right, value = cast_datetime(left), cast_datetime(right), cast_datetime(value)
            elif type == "date": left, right, value = cast_date(left), cast_date(right), cast_date(value)
            data = self.filter_partition(data, field, left, right, value)
        return sorted(data[field].unique())

    def filter_partition(self, data: pd.DataFrame, field: str, left=None, right=None, value=None) -> pd.DataFrame:
        __left = (data[field] >= left) if left is not None else data[field].notna()
        __right = (data[field] <= right) if right is not None else data[field].notna()
        __value = (data[field] == value) if value is not None else data[field].notna()
        return data[__left&__right&__value]

    def __iter__(self):
        return self

    def __next__(self):
        if self.index < len(self.partition):
            partition = self.data[self.data[self.field]==self.partition[self.index]]
            self.index += 1
            return partition.drop(columns=[self.field]) if self.drop_partition else partition
        else: raise StopIteration

    def __len__(self):
        return len(self.partition)


def create_connection(project_id: str, account: Account=dict()) -> bigquery.Client:
    account = account if account and isinstance(account, dict) else read_gcloud(account)
    return bigquery.Client.from_service_account_info(account, project=project_id)


def execute_query(query: str, project_id=str(), account: Account=dict()) -> RowIterator:
    client = create_connection(project_id, account)
    job = client.query(query)
    return job.result()


def make_select_query(table: str, columns: IndexLabel=list(), field=str(), type: Literal["date","datetime","number"]="date",
                    left=None, right=None, value=None, limited=True, has_value=True, **kwargs) -> str:
    query = f"SELECT {', '.join(columns) if columns else '*'} FROM `{table}`"
    if not field: return query + ';'
    elif (not limited) or has_value:
        return query + ' ' + get_partition_filter(field, type, left, right, value) + ';'
    else: raise ValueError(INVALID_MSG(PARTITION_FILTER))


def make_delete_query(table: str, field=str(), type: Literal["date","datetime","number"]="date",
                    left=None, right=None, value=None, limited=True, has_value=True, **kwargs) -> str:
    query = f"DELETE FROM `{table}`"
    if not field: return query + ' ' + "WHERE TRUE" + ';'
    elif (not limited) or has_value:
        return query + ' ' + get_partition_filter(field, type, left, right, value) + ';'
    else: raise ValueError(INVALID_MSG(PARTITION_FILTER))


def get_partition_filter(field=str(), type: Literal["date","datetime","number"]="date", left=None, right=None, value=None, **kwargs) -> str:
    if type in ("date","datetime"):
        return _get_datetime_partition_filter(field, type, left, right, value)
    elif value is not None:
        return f"WHERE {field} = {value}"
    elif (left is not None) and (right is not None):
        if left == right: return f"WHERE {field} = {left}"
        else: return f"WHERE {field} BETWEEN {left} AND {right}"
    elif left is not None:
        return f"WHERE {field} >= {left}"
    elif right is not None:
        return f"WHERE {field} <= {right}"
    else: return "WHERE TRUE"


def _get_datetime_partition_filter(field=str(), dtype: Literal["date","datetime"]="date", left=None, right=None, value=None) -> str:
    def set_filed(field: str, dtype: Literal["date","datetime"], value: dt.date) -> str:
        return f"DATE({field})" if (dtype == "datetime") and not isinstance(value, dt.datetime) else field
    if isinstance(value, dt.date):
        return f"WHERE {set_filed(field, dtype, value)} = '{value}'"
    elif isinstance(left, dt.date) and isinstance(right, dt.date) and (type(left) == type(right)):
        field = set_filed(field, dtype, left)
        if left == right: return f"WHERE {field} = '{left}'"
        else: return f"WHERE {field} BETWEEN '{left}' AND '{right}'"
    elif isinstance(left, dt.date):
        return f"WHERE {set_filed(field, dtype, left)} >= '{left}'"
    elif isinstance(right, dt.date):
        return f"WHERE {set_filed(field, dtype, right)} <= '{right}'"
    else: return "WHERE TRUE"


def read_gbq(query: str, project_id: str, return_type: Literal["records","dataframe"]="dataframe",
            account: Account=dict()) -> TabularData:
    client = create_connection(project_id, account)
    query_job = client.query(query if query.upper().startswith("SELECT") else f"SELECT * FROM `{query}`;")
    if return_type == "dataframe": return query_job.to_dataframe()
    else: return [dict(row.items()) for row in query_job.result()]


def upload_gbq(table: str, project_id: str, data: pd.DataFrame, if_exists: Literal["replace","append"]="append",
                partition: BigQueryPartition=dict(), progress=True, account: Account=dict()):
    client = create_connection(project_id, account)
    if if_exists == "replace":
        client.query(make_delete_query(table, **partition))
    job_config = LoadJobConfig(write_disposition="WRITE_APPEND")
    iterator = PartitionedIterator(data, **partition) if _is_partitioned_data(data, partition) else [data]
    for __data in tqdm(iterator, desc=UPLOAD_GBQ_MSG(table), disable=(not progress)):
        client.load_table_from_dataframe(__data, f"{project_id}.{table}", job_config=job_config)


def _is_partitioned_data(data: pd.DataFrame, partition: BigQueryPartition=dict()) -> bool:
    return isinstance(partition, Dict) and partition.get("field") and (partition["field"] in data)


def upsert_gbq(table: str, project_id: str, data: pd.DataFrame, primary_key: List[str],
                base: Optional[pd.DataFrame]=None, account: Account=dict()):
    if df_empty(base):
        base = read_gbq(table, project_id, return_type="dataframe", account=account)
    data = data.set_index(primary_key).combine_first(base.set_index(primary_key)).reset_index()
    client = create_connection(project_id, account)
    client.query(f"DELETE FROM `{table}` WHERE TRUE;")
    job_config = LoadJobConfig(write_disposition="WRITE_APPEND")
    client.load_table_from_dataframe(data, f"{project_id}.{table}", job_config=job_config)



from __future__ import annotations
from gsdriver.base.abstract import SESSION_CONTEXT
from gsdriver.base.session import Info, Schema, Field, Flow, Process
from gsdriver.base.session import ITER_INDEX, PAGE_ITERATOR, iter_task
from gsdriver.base.gcloud import GoogleQueryList, GoogleUploadList

from gsdriver.base.spider import Spider, AsyncSpider, EncryptedSpider, EncryptedAsyncSpider
from gsdriver.base.spider import LoginSpider, LoginCookie, RangeFilter, SPIDER_KILL, ENCRYPTED_SPIDER_KILL
from gsdriver.base.spider import MAX_ASYNC_TASK_LIMIT, MAX_REDIRECT_LIMIT, RETRY, RETURN
from gsdriver.base.spider import ACTION, WHERE, WHICH, DRIVER_UNIQUE

from gsdriver.base.types import LogLevel, TypeHint, EncryptedKey, DecryptedKey
from gsdriver.base.types import IndexLabel, Keyword, Unit, Range, Timedelta, Timezone, JsonData

from gsdriver.utils.alert import AlertInfo
from gsdriver.utils.logs import log_object, log_messages
from gsdriver.utils.map import notna_dict, search_keyword, get_scala
from gsdriver.utils.request import get_headers, encode_params

from webdriver_manager.chrome import ChromeDriverManager
from selenium import webdriver
from seleniumwire import webdriver as webdriver_wire
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.chrome.options import Options
from seleniumwire.request import Request

from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.remote.webelement import WebElement

import abc
import asyncio
import functools
import pyperclip
import requests
import time
import os
import subprocess

from typing import Any, Callable, Dict, List, Literal, Optional, Sequence, Tuple, Union
from numbers import Real
from bs4 import BeautifulSoup, Tag
from gzip import GzipFile
from io import BytesIO
import json
import re


SOURCE, BYTES, TEXT, JSON, HTML = "source", "bytes", "text", "json", "html"
REFRESH, REOPEN = 2, 3

class DriverError(NameError):
    def __init__(self, enableWire=False):
        message = f"Selenium{'-Wire' if enableWire else ''} WebDriver does not exist."
        super().__init__(message)

SELENIUM_SPIDER_KILL = (*SPIDER_KILL, DriverError)
SELENIUM_ENCRYPTED_SPIDER_KILL = (*ENCRYPTED_SPIDER_KILL, DriverError)


###################################################################
############################# Selenium ############################
###################################################################

def wait_element(func):
    @functools.wraps(func)
    def wrapper(element: WebElement, *args, delay=0.1, retry=0, **kwargs):
        for _ in range(retry):
            try: return func(element, *args, **kwargs)
            except NoSuchElementException:
                if delay: time.sleep(delay)
        return func(element, *args, **kwargs)
    return wrapper


@wait_element
def select_element(element: WebElement, selector: str, delay=0.1, retry=0) -> WebElement:
    return element.find_element(By.CSS_SELECTOR, selector)


@wait_element
def select_elements(element: WebElement, selector: str, if_null: Literal["pass","error"]="pass",
                    delay=0.1, retry=0) -> List[WebElement]:
    elements = element.find_elements(By.CSS_SELECTOR, selector)
    if (if_null == "error") and (len(elements) == 0):
        _raise_no_such_element_exception(selector)
    else: return elements


@wait_element
def click_text(element: WebElement, selector: str, value: str, if_null: Literal["pass","error"]="pass",
                exact=True, delay=0.1, retry=0):
    elements = element.find_elements(By.CSS_SELECTOR, selector)
    for __element in elements:
        if (__element.text.strip() == value) if exact else (value in __element.text):
            return __element.click()
    if if_null == "error":
        _raise_no_such_element_exception(selector)


def _raise_no_such_element_exception(selector: str):
    context = '{"method":"css selector","selector":"'+selector+'"}'
    raise NoSuchElementException("Message: no such element: Unable to locate element: "+context)


@wait_element
def input_text(element: WebElement, selector: str, value: str, clear=True, delay=0.1, retry=0):
    input = element.find_element(By.CSS_SELECTOR, selector)
    if clear: input.clear()
    input.send_keys(value)


@wait_element
def paste_text(element: WebElement, selector: str, value: str, delay=0.1, retry=0):
    input = element.find_element(By.CSS_SELECTOR, selector)
    input.click()
    pyperclip.copy(value)
    input.send_keys(Keys.CONTROL, 'v')


def wait_exists(element: WebElement, selector: str, delay=0.1):
    def select_element(element: WebElement, selector: str) -> bool:
        try: return isinstance(element.find_element(By.CSS_SELECTOR, selector), WebElement)
        except: return False
    while select_element(element, selector):
        time.sleep(delay)


def move_to_element(driver: webdriver.Chrome, element: Optional[WebElement]=None, selector=str(), delay=0.1, retry=0):
    actions = ActionChains(driver)
    if selector:
        element = select_element(driver, selector, delay=delay, retry=retry)
    actions.move_to_element(element).perform()


def scroll_to_bottom(driver: webdriver.Chrome):
    driver.execute_script("window.scrollTo(0, document.body.scrollHeight);")


def refresh_hard(driver: webdriver.Chrome):
    driver.execute_script("location.reload(true);")


def open_window(driver: webdriver.Chrome):
    driver.execute_script("window.open('');")


def switch_window(driver: webdriver.Chrome, index: int):
    driver.switch_to.window(driver.window_handles[index])


###################################################################
########################## Selenium-Wire ##########################
###################################################################

ResponseBody = Union[bytes, str, JsonData, Tag]
ResponseData = Union[ResponseBody, List[ResponseBody]]

RequestAttr = Union[Request, Dict[str,str], ResponseBody]
RequestData = Union[RequestAttr, List[RequestAttr]]

class NoSuchRequestException(NoSuchElementException):
    ...


def wait_request(func):
    @functools.wraps(func)
    def wrapper(driver: webdriver_wire.Chrome, *args, empty=True, delay=0.1, retry=0, **kwargs):
        for _ in range(retry):
            try: return func(driver, *args, empty=False, **kwargs)
            except NoSuchRequestException:
                if delay: time.sleep(delay)
        return func(driver, *args, empty=empty, **kwargs)
    return wrapper


def clear_request(func):
    @functools.wraps(func)
    def wrapper(driver: webdriver_wire.Chrome, *args, clear_requests=False, **kwargs):
        results = func(driver, *args, **kwargs)
        if clear_requests: del driver.requests
        return results
    return wrapper


def parse_safe(func):
    @functools.wraps(func)
    def wrapper(*args, safe=True, default=None, **kwargs):
        if safe:
            try: return func(*args, **kwargs)
            except: return default
        else: return func(*args, **kwargs)
    return wrapper


@wait_request
@clear_request
def get_requests(driver: webdriver_wire.Chrome, match: Dict[str,Any]=dict(), how: Literal["and","or"]="and", index: Optional[int]=None, attr=None,
                parse: Optional[Literal["bytes","text","json","html"]]=None, decoder: Optional[Callable[[bytes],bytes]]=None,
                safe=True, with_index=False, with_url=False, empty=True, delay=0.1, retry=0, clear_requests=False) -> RequestData:
    results = list()
    for __index, __request in enumerate(driver.requests):
        if _match_request(__request, match, how, parse=parse, decoder=decoder, safe=safe):
            __result = parse_request(__request, attr, parse=parse, decoder=decoder, safe=safe)
            if __result is not None:
                if with_index: results.append((__index, __result))
                elif with_url: results.append((__request.url, __result))
                else: results.append(__result)
    if not (empty or results):
        raise NoSuchRequestException("Message: no such request")
    elif isinstance(index, int):
        return get_scala(results, index, default=((None, None) if with_index or with_url else None))
    else: return results


def _match_request(request: Request, match: Dict[str,Any], how: Literal["and","or"]="and", safe=True, **kwargs) -> bool:
    if not match: return True
    elif how == "and": return _match_request_and(request, match, safe=safe, **kwargs)
    elif how == "or": return _match_request_or(request, match, safe=safe, **kwargs)
    else: return False


def _match_request_and(request: Request, match: Dict[str,Any], safe=True, **kwargs) -> bool:
    __match = True
    for __attr, __func in match.items():
        try:
            if __attr == "response":
                for __attr, __func in (__func if isinstance(__func, Dict) else dict()).items():
                    if __attr == "body": __match &= _match_response_body(request, __func, safe=safe, default=False, **kwargs)
                    else: __match &= _match_attr(getattr(request.response, __attr), __func, safe=safe, default=False)
            else: __match &= _match_attr(getattr(request, __attr), __func, safe=safe, default=False)
        except: return False
    return __match


def _match_request_or(request: Request, match: Dict[str,Any], safe=True, **kwargs) -> bool:
    for __attr, __func in match.items():
        try:
            if __attr == "response":
                for __attr, __func in (__func if isinstance(__func, Dict) else dict()).items():
                    if __attr == "body":
                        if _match_response_body(request, __func, safe=safe, default=False, **kwargs): return True
                        else: pass
                    elif _match_attr(getattr(request.response, __attr), __func, safe=safe, default=False): return True
                    else: pass
            elif _match_attr(getattr(request, __attr), __func, safe=safe, default=False): return True
            else: pass
        except: pass
    return False


@parse_safe
def _match_attr(attr: Any, match: Any, safe=True, default=None) -> bool:
    if isinstance(match, Callable): return match(attr)
    elif isinstance(match, re.Pattern): return bool(match.search(_decode_attr(attr)))
    elif isinstance(match, Sequence): return search_keyword(_decode_attr(attr), match)
    else: return False


@parse_safe
def _match_response_body(request: Request, match: Any, parse: Optional[Literal["bytes","text","json","html"]]=None,
                    decoder: Optional[Callable[[bytes],bytes]]=None, safe=True, default=None) -> bool:
    if isinstance(match, Callable):
        return match(parse_request(request, None, parse, decoder, safe=False))
    elif isinstance(match, re.Pattern):
        return bool(match.search(parse_request(request, parse=TEXT, decoder=decoder, safe=False, default=str())))
    elif isinstance(match, Sequence):
        return search_keyword(parse_request(request, parse=TEXT, decoder=decoder), match, safe=False, default=str())
    else: return False


@parse_safe
def parse_request(request: Request, attr=None, parse: Optional[Literal["bytes","text","json","html"]]=None,
                decoder: Optional[Callable[[bytes],bytes]]=None, safe=True, default=None) -> RequestAttr:
    if attr or parse:
        __attr = getattr(request, attr) if attr else _get_response_body(request, decoder)
        if parse == BYTES: return _encode_attr(__attr)
        elif parse == TEXT: return _decode_attr(__attr)
        elif parse == JSON: return json.loads(__attr)
        elif parse == HTML: return BeautifulSoup(__attr, "html.parser")
        else: return __attr
    else: return request


def _get_response_body(request: Request, decoder: Optional[Callable[[bytes],bytes]]=None) -> bytes:
    decoder = gzip_decoder if isinstance(decoder, str) and decoder == "gzip" else decoder
    return decoder(request.response.body) if isinstance(decoder, Callable) else request.response.body


def _encode_attr(attr: Any) -> bytes:
    if isinstance(attr, bytes): return attr
    elif isinstance(attr, (List,Dict)):
        return json.dumps(attr, ensure_ascii=False).encode("utf-8")
    else: return str(attr).encode("utf-8")


def _decode_attr(attr: Any) -> str:
    return attr.decode("utf-8") if isinstance(attr, bytes) else str(attr)


def gzip_decoder(body: bytes) -> bytes:
    with GzipFile(fileobj=BytesIO(body)) as f:
        return f.read()


###################################################################
########################## Selenium Base ##########################
###################################################################

Port = Union[str,int]

class SeleniumBase(webdriver.Chrome):
    process = None

    def set_service(self, chromePath=str(), debuggingPort: Optional[Port]=None, debuggingDir=str(),
                    proxyPort: Optional[Port]=None, incognito=False) -> Service:
        if chromePath:
            if self.validate_port(debuggingPort) and debuggingDir:
                self.run_debugging(chromePath, debuggingPort, debuggingDir, proxyPort, incognito)
            else: return Service(executable_path=chromePath)
        try: return Service(executable_path=ChromeDriverManager().install())
        except: return Service(executable_path=r".\chromedriver.exe")

    def set_options(self, downloadPath=str(), headless=False, incognito=False, loadImages=True, disableProxy=False, ignoreCert=False,
                    extensions: List[str]=list(), debuggingPort: Optional[Port]=None, proxyPort: Optional[Port]=None) -> Options:
        options = Options()
        options.add_argument("--disable-infobars")
        options.add_argument("--disable-blink-features=AutomationControlled")
        if headless: self._add_headless_options(options)
        if incognito: options.add_argument("--incognito")
        if disableProxy: options.add_argument("--disable-proxy-certificate-handler")
        if ignoreCert: options.add_argument("--ignore-certificate-errors")
        if extensions: self._add_extensions(options, extensions)
        if self.validate_port(proxyPort): options.add_argument(f"--proxy-server=http://localhost:{proxyPort}")
        self._add_experimental_options(options, downloadPath, loadImages, debuggingPort)
        return options

    def _add_headless_options(self, options: Options, gpu=False, windowSize="1920x1080"):
        options.add_argument("--headless")
        if not gpu: options.add_argument("--disable-gpu")
        if windowSize: options.add_argument(f"--window-size={windowSize}")
        options.add_argument("user-agent="+get_headers()["User-Agent"])

    def _add_extensions(self, options: Options, extensions: List[str]):
        for __extension in extensions:
            if os.path.exists(__extension): options.add_extension(__extension)

    def _add_experimental_options(self, options: Options, downloadPath=str(), loadImages=True, debuggingPort: Optional[Port]=None):
        prefs = dict()
        if downloadPath:
            prefs["download.default_directory"] = os.path.abspath(downloadPath)
            prefs["download.prompt_for_download"] = False
            prefs["download.directory_upgrade"] = True
        if not loadImages: prefs["profile.managed_default_content_settings.images"] = 2
        if prefs: options.add_experimental_option("prefs", prefs)
        if self.validate_port(debuggingPort): options.add_experimental_option("debuggerAddress", f"127.0.0.1:{debuggingPort}")

    def set_timeout_and_wait(self, loadTimeout: Optional[float]=None, implicitlyWait: Optional[float]=None):
        if isinstance(loadTimeout, (float,int)) and (loadTimeout > 0.):
            self.set_page_load_timeout(loadTimeout)
        if isinstance(implicitlyWait, (float,int)) and (implicitlyWait > 0.):
            self.implicitly_wait(implicitlyWait)

    ###################################################################
    ########################### Debug Driver ##########################
    ###################################################################

    def run_debugging(self, chromePath: str, debuggingPort: int, debuggingDir: str, proxyPort: Optional[Port]=None, incognito=False) -> Service:
        command = f'"{chromePath}" --remote-debugging-port={debuggingPort} --user-data-dir="{debuggingDir}"'
        if self.validate_port(proxyPort): command += f" --proxy-server=localhost:{proxyPort}"
        if incognito: command = command + " --incognito"
        self.process = subprocess.Popen(command)

    def exit_debugging(self):
        try:
            if isinstance(self.process, subprocess.Popen) and (self.process.poll() is None):
                self.process.terminate()
        except: return

    def validate_port(self, port: Optional[Port]=None) -> bool:
        if port is None: return False
        elif isinstance(port, int): return True
        else: return str(port).isdecimal()

    ###################################################################
    ########################## Request Driver #########################
    ###################################################################

    def catch_exception(func):
        @functools.wraps(func)
        def wrapper(self: SeleniumBase, *args, ignore_exception=False, **context):
            if ignore_exception:
                try: return func(self, *args, **context)
                except: return None
            else: return func(self, *args, **context)
        return wrapper

    @catch_exception
    def get(self, url: str, params=None, headers=None, parse: Optional[Literal["text","json","html"]]=None,
            delay: Optional[Real]=None, ignore_exception=False) -> Union[str,JsonData,Tag]:
        super().get(encode_params(url, params) if isinstance(params, (Dict,Sequence)) and params else url)
        if isinstance(delay, (float,int)) and delay:
            time.sleep(delay)
        return self.get_page_source(parse)

    @catch_exception
    def get_page_source(self, parse: Optional[Literal["text","json","html"]]=None, ignore_exception=False) -> Union[str,JsonData,Tag]:
        if parse == TEXT: return BeautifulSoup(self.page_source, "html.parser").text
        elif parse == JSON: return json.loads(BeautifulSoup(self.page_source, "html.parser").text)
        elif parse == HTML: return BeautifulSoup(self.page_source, "html.parser")
        else: return self.page_source

    ###################################################################
    ######################### Selenium Method #########################
    ###################################################################

    def select_element(self, selector: str, delay=0.1, retry=0) -> WebElement:
        return select_element(self, selector, delay=delay, retry=retry)

    def select_elements(self, selector: str, if_null: Literal["pass","error"]="pass",
                        delay=0.1, retry=0) -> List[WebElement]:
        return select_elements(self, selector, if_null=if_null, delay=delay, retry=retry)

    def click_text(self, selector: str, value: str, if_null: Literal["pass","error"]="pass",
                    exact=True, delay=0.1, retry=0):
        click_text(self, selector, value, if_null=if_null, exact=exact, delay=delay, retry=retry)

    def input_text(self, selector: str, value: str, clear=True, delay=0.1, retry=0):
        input_text(self, selector, value, clear=clear, delay=delay, retry=retry)

    def paste_text(self, selector: str, value: str, delay=0.1, retry=0):
        paste_text(self, selector, value, delay=delay, retry=retry)

    def wait_exists(self, selector: str, delay=0.1):
        wait_exists(self, selector, delay)

    def move_to_element(self, element: Optional[WebElement]=None, selector=str(), delay=0.1, retry=0):
        move_to_element(self, element, selector, delay=delay, retry=retry)

    def scroll_to_bottom(self, delay=0.):
        scroll_to_bottom(self)
        if delay: time.sleep(delay)

    def refresh_hard(self, delay=0.):
        refresh_hard(self)
        if delay: time.sleep(delay)

    def open_window(self, delay=0.):
        open_window(self)
        if delay: time.sleep(delay)

    def switch_window(self, index: int, delay=0.):
        switch_window(self, index)
        if delay: time.sleep(delay)


###################################################################
######################### Selenium Driver #########################
###################################################################

class SeleniumDriver(SeleniumBase):
    def __init__(self, chromePath=str(), downloadPath=str(), headless=False, incognito=False, loadImages=True, disableProxy=False,
                ignoreCert=False, loadTimeout: Optional[float]=None, implicitlyWait: Optional[float]=None, extensions: List[str]=list(),
                debuggingPort: Optional[Port]=None, debuggingDir=str(), proxyPort: Optional[Port]=None, **kwargs):
        service = self.set_service(chromePath, debuggingPort, debuggingDir, proxyPort, incognito)
        options = self.set_options(downloadPath, headless, incognito, loadImages, disableProxy, ignoreCert, extensions, debuggingPort, proxyPort)
        super().__init__(service=service, options=options)
        self.set_timeout_and_wait(loadTimeout, implicitlyWait)

    def __exit__(self, exc_type, exe, traceback):
        super().__exit__(exc_type, exe, traceback) # quit()
        self.exit_debugging()

    def close(self):
        super().close()
        self.exit_debugging()

    def quit(self):
        super().quit()
        self.exit_debugging()

    def get_cookies(self, encode=True, raw=False) -> Union[str,Dict,List]:
        if raw: return super().get_cookies()
        elif encode: return '; '.join([f'{cookie["name"]}={cookie["value"]}' for cookie in super().get_cookies()])
        else: return {cookie["name"]: cookie["value"] for cookie in super().get_cookies()}

    def set_cookies(self, session: requests.Session):
        for cookie in self.get_cookies(raw=True):
            kwargs = dict(domain=cookie.get("domain"), path=cookie.get("path"), secure=cookie.get("secure",False), expires=cookie.get("expiry"))
            session.cookies.set(cookie["name"], cookie["value"], **kwargs)


###################################################################
########################## Selenium Wire ##########################
###################################################################

class SeleniumWireDriver(webdriver_wire.Chrome, SeleniumBase):
    def __init__(self, chromePath=str(), downloadPath=str(), headless=False, incognito=False, loadImages=True, disableProxy=False,
                ignoreCert=False, loadTimeout: Optional[float]=None, implicitlyWait: Optional[float]=None, extensions: List[str]=list(),
                debuggingPort: Optional[Port]=None, debuggingDir=str(), proxyPort: Optional[Port]=None, **kwargs):
        service = self.set_service(chromePath, debuggingPort, debuggingDir, proxyPort, incognito)
        options = self.set_options(downloadPath, headless, incognito, loadImages, disableProxy, ignoreCert, extensions, debuggingPort, proxyPort)
        seleniumwire_options = dict(port=int(proxyPort)) if self.validate_port(proxyPort) else None
        webdriver_wire.Chrome.__init__(self, service=service, options=options, seleniumwire_options=seleniumwire_options)
        self.set_timeout_and_wait(loadTimeout, implicitlyWait)

    def __exit__(self, exc_type, exe, traceback):
        webdriver_wire.Chrome.__exit__(self, exc_type, exe, traceback) # quit()
        self.exit_debugging()

    def close(self):
        webdriver_wire.Chrome.close(self)
        self.exit_debugging()

    def quit(self):
        webdriver_wire.Chrome.quit(self)
        self.exit_debugging()

    @SeleniumBase.catch_exception
    def get(self, url: str, params=None, headers=None, parse: Optional[Literal["text","json","html"]]=None,
                delay: Optional[Real]=None, ignore_exception=False) -> Union[str,JsonData,Tag]:
        if isinstance(headers, Dict):
            self.set_headers(headers)
        webdriver_wire.Chrome.get(self, encode_params(url, params) if isinstance(params, (Dict,Sequence)) and params else url)
        if isinstance(delay, (float,int)) and delay:
            time.sleep(delay)
        return self.get_page_source(parse)

    def get_cookies(self, encode=True, raw=False) -> Union[str,Dict,List]:
        if raw: return super().get_cookies()
        elif encode: return '; '.join([cookie["name"]+'='+cookie["value"] for cookie in super().get_cookies()])
        else: return {cookie["name"]: cookie["value"] for cookie in super().get_cookies()}

    def set_cookies(self, session: requests.Session):
        for cookie in self.get_cookies(raw=True):
            kwargs = dict(domain=cookie.get("domain"), path=cookie.get("path"), secure=cookie.get("secure",False), expires=cookie.get("expiry"))
            session.cookies.set(cookie["name"], cookie["value"], **kwargs)

    def set_headers(self, headers: Dict[str,str]):
        def interceptor(request: Request):
            for __key, __value in headers.items():
                request.headers[__key] = __value
        self.request_interceptor = interceptor

    def get_requests(self, match: Dict[str,Any]=dict(), how: Literal["and","or"]="and", index: Optional[int]=None, attr=None,
                    parse: Optional[Literal["bytes","text","json","html"]]=None, decoder: Optional[Callable[[bytes],bytes]]=None,
                    safe=True, with_index=False, with_url=False, empty=True, delay=0.1, retry=0, clear_requests=False) -> RequestData:
        kwargs = dict(empty=empty, delay=delay, retry=retry, clear_requests=clear_requests)
        return get_requests(self, match, how, index, attr, parse, decoder, safe, with_index, with_url, **kwargs)


###################################################################
######################### Selenium Spider #########################
###################################################################

ChromeDriver = Union[SeleniumDriver,SeleniumWireDriver]

class SeleniumSpider(Spider):
    __metaclass__ = abc.ABCMeta
    asyncio = False
    operation = "spider"
    host = str()
    where = WHERE
    which = WHICH
    verb = ACTION
    by = str()
    fields = list()
    ranges = list()
    iterateArgs = list()
    iterateCount = dict()
    iterateProduct = list()
    iterateUnit = 1
    redirectUnit = 1
    pagination = False
    pageFrom = 1
    offsetFrom = 1
    pageUnit = 0
    pageLimit = 0
    interval = None
    fromNow = None
    timeout = None
    ssl = None
    numRetries = 0
    delay = 1.
    interruptType = tuple()
    killType = tuple()
    errorType = tuple()
    responseType = "records"
    returnType = "records"
    mappedReturn = False
    root = list()
    groupby = list()
    groupSize = dict()
    countby = str()
    driver = None
    enableWire = False
    info = Info()
    flow = Flow()

    def __init__(self, fields: Optional[IndexLabel]=None, ranges: Optional[RangeFilter]=None, returnType: Optional[TypeHint]=None,
                tzinfo: Optional[Timezone]=None, countryCode=str(), datetimeUnit: Optional[Literal["second","minute","hour","day"]]=None,
                logName: Optional[str]=None, logLevel: LogLevel="WARN", logFile: Optional[str]=None, localSave=False,
                debugPoint: Optional[Keyword]=None, killPoint: Optional[Keyword]=None, extraSave: Optional[Keyword]=None,
                numRetries: Optional[int]=None, delay: Range=1., cookies: Optional[str]=None,
                fromNow: Optional[Unit]=None, discard=True, progress=True, where=str(), which=str(), by=str(), message=str(),
                iterateUnit: Optional[int]=None, interval: Optional[Timedelta]=None, apiRedirect=False, redirectUnit: Optional[int]=None,
                queryList: GoogleQueryList=list(), uploadList: GoogleUploadList=list(), alertInfo: AlertInfo=dict(),
                driverOptions: Optional[Dict]=None, reopenDelay: Range=30., **context):
        Spider.__init__(self, **self.from_locals(locals(), drop=DRIVER_UNIQUE))
        self.set_driver_options(driverOptions, reopenDelay)

    def find_driver(self, driver: Optional[ChromeDriver]=None, **context) -> ChromeDriver:
        __class = SeleniumWireDriver if self.enableWire else SeleniumDriver
        if isinstance(driver, __class): return driver
        elif isinstance(self.driver, __class): return self.driver
        else: raise DriverError(self.enableWire)

    def get_driver_options(self, driverOptions: Optional[Dict]=None, **context) -> Dict:
        return driverOptions if isinstance(driverOptions, Dict) else dict()

    def set_driver_options(self, driverOptions: Optional[Dict]=None, reopenDelay: Range=30.):
        if isinstance(driverOptions, Dict) and driverOptions:
            options = dict(driverOptions=driverOptions, reopenDelay=reopenDelay)
        else: options = dict(reopenDelay=reopenDelay)
        self.update(options)

    ###################################################################
    ######################### Driver Managers #########################
    ###################################################################

    def init_driver(func):
        @functools.wraps(func)
        def wrapper(self: SeleniumSpider, *args, self_var=True, **context):
            args, context = self.init_context(args, context, self_var=self_var)
            self.run_driver(**self.get_driver_options(**context))
            try: data = func(self, *args, **SESSION_CONTEXT(**context))
            finally: self.quit_driver()
            self.with_data(data, func=func.__name__, **context)
            return data
        return wrapper

    def with_driver(func):
        @functools.wraps(func)
        def wrapper(self: SeleniumSpider, *args, **context):
            self.run_driver(**self.get_driver_options(**context))
            try: data = func(self, *args, **context)
            finally: self.quit_driver()
            return data
        return wrapper

    def is_kill(self, exception: Exception) -> bool:
        return isinstance(exception, SELENIUM_SPIDER_KILL) or isinstance(exception, self.killType)

    ###################################################################
    ############################ Run Driver ###########################
    ###################################################################

    def run_driver(self, **options):
        self.driver = SeleniumWireDriver(**options) if self.enableWire else SeleniumDriver(**options)

    @Spider.ignore_exception
    def quit_driver(self, clear=False, delay=0.25):
        if clear: self.driver.delete_all_cookies()
        self.driver.quit()
        self.sleep(delay)

    def reopen_driver(self, driverOptions: Dict=dict(), reopenDelay: Range=30.):
        self.quit_driver(clear=True, delay=0.)
        self.sleep(reopenDelay, minimum=1.)
        self.run_driver(**driverOptions)

    @Spider.ignore_exception
    def refresh_driver(self, how: Literal["normal","hard"]="normal", delay: Optional[Range]=None):
        if how == "hard": self.driver.refresh_hard()
        else: self.driver.refresh()
        self.sleep(delay)

    def interrupt(self, *args, exception: Exception, func: Callable, retryCount=0, reopenDelay: Range=30., **context) -> Tuple[int,Any]:
        if self.driver is not None:
            self.reopen_driver(self.get_driver_options(**context), reopenDelay)
            return super().interrupt(*args, exception=exception, func=func, retryCount=retryCount, **context)
        else: raise exception

    def interrupt_by_count(self, numRetries=0, retryCount=0, reopenDelay=30., delay: Optional[Range]=None,
                            refresh: Literal["auto","normal","hard"]="auto", **context) -> int:
        if (numRetries > 1) and (retryCount == 0):
            self.reopen_driver(self.get_driver_options(**context), reopenDelay)
            return REOPEN
        elif (numRetries != retryCount) and (retryCount > 0):
            if refresh == "auto":
                refresh = "normal" if ((numRetries-1) == retryCount) and (retryCount > 0) else "hard"
            self.refresh_driver(delay=delay, how=refresh)
            return REFRESH
        else: RETURN

    ###################################################################
    ########################## Request Driver #########################
    ###################################################################

    def driver_required(func):
        @functools.wraps(func)
        def wrapper(self: SeleniumSpider, *args, driver: Optional[ChromeDriver]=None, **context):
            return func(self, *args, driver=self.find_driver(driver), **context)
        return wrapper

    @driver_required
    def request_driver(self, how: Literal["source","text","json","html"], url: str, params=None, headers=None,
                        delay: Optional[Range]=None, ignore_exception=False, *args,
                        driver: ChromeDriver, **context) -> Union[str,JsonData,Tag]:
        messages = notna_dict(params=params, headers=headers)
        self.logger.debug(log_messages(**notna_dict({ITER_INDEX: context.get(ITER_INDEX)}), **messages))
        self.checkpoint(iter_task(context, "request"), where="request_driver", msg=dict(url=url, **messages))
        response = driver.get(url, **messages, parse=None, delay=self.get_delay(delay), ignore_exception=ignore_exception)
        self.logger.info(log_object(dict(self.get_iterator(**context, _index=True), **{"url":url, "contents-length":len(response)})))
        self.checkpoint(iter_task(context, "response"), where="request_driver", msg={"response":response}, save=response)
        return driver.get_page_source(how, ignore_exception=ignore_exception) if how != SOURCE else response

    @driver_required
    def get_page_source(self, parse: Optional[Literal["text","json","html"]]=None, ignore_exception=False, *args,
                        driver: ChromeDriver, **context) -> Union[str,JsonData,Tag]:
        return driver.get_page_source(parse, ignore_exception=ignore_exception)

    ###################################################################
    ########################## Selenium Wire ##########################
    ###################################################################

    def wire_required(func):
        @functools.wraps(func)
        def wrapper(self: SeleniumSpider, *args, driver: Optional[SeleniumWireDriver]=None, **context):
            if self.enableWire:
                return func(self, *args, driver=self.find_driver(driver), **context)
            else: raise DriverError(enableWire=True)
        return wrapper

    @wire_required
    def get_driver_response(self, match: Dict[str,Any]=dict(), how: Literal["and","or"]="and", index: Optional[int]=None,
                            parse: Optional[Literal["bytes","text","json","html"]]=None, decoder: Optional[Callable[[bytes],bytes]]=None,
                            safe=True, empty=True, delay=0.1, retry=0, clear_requests=False, *args,
                            driver: Optional[SeleniumWireDriver]=None, **context) -> ResponseData:
        kwargs = dict(empty=empty, delay=delay, retry=retry, clear_requests=clear_requests)
        response = driver.get_requests(match, how, index, None, parse, decoder, safe, with_url=True, **kwargs)
        for __url, __response in ([response] if isinstance(index, int) else response):
            self.checkpoint(iter_task(context, "driver"), where="get_response", msg={"response":__response}, save=__response)
            self.logger.debug(log_messages(url=__url, match=match))
        return response[1] if isinstance(index, int) else [__response[1] for __response in response]

    ###################################################################
    ######################### Selenium Method #########################
    ###################################################################

    def select_element(self, element: WebElement, selector: str, delay=0.1, retry=0) -> WebElement:
        return select_element(element, selector, delay=delay, retry=retry)

    def select_elements(self, element: WebElement, selector: str, if_null: Literal["pass","error"]="pass",
                        delay=0.1, retry=0) -> List[WebElement]:
        return select_elements(element, selector, if_null=if_null, delay=delay, retry=retry)

    def click_text(self, element: WebElement, selector: str, value: str, if_null: Literal["pass","error"]="pass",
                    exact=True, delay=0.1, retry=0):
        click_text(element, selector, value, if_null=if_null, exact=exact, delay=delay, retry=retry)

    def input_text(self, element: WebElement, selector: str, value: str, clear=True, delay=0.1, retry=0):
        input_text(element, selector, value, clear=clear, delay=delay, retry=retry)

    def paste_text(self, element: WebElement, selector: str, value: str, delay=0.1, retry=0):
        paste_text(element, selector, value, delay=delay, retry=retry)

    def wait_exists(self, element: WebElement, selector: str, delay=0.1):
        wait_exists(element, selector, delay)

    def move_to_element(self, driver: webdriver.Chrome, element: Optional[WebElement]=None, selector=str(), delay=0.1, retry=0):
        move_to_element(driver, element, selector, delay=delay, retry=retry)


###################################################################
###################### Selenium Async Spider ######################
###################################################################

class SeleniumAsyncSpider(AsyncSpider, SeleniumSpider):
    __metaclass__ = abc.ABCMeta
    asyncio = True
    operation = "spider"
    host = str()
    where = WHERE
    which = WHICH
    verb = ACTION
    by = str()
    fields = list()
    ranges = list()
    iterateArgs = list()
    iterateCount = dict()
    iterateProduct = list()
    iterateUnit = 1
    maxLimit = MAX_ASYNC_TASK_LIMIT
    redirectUnit = 1
    redirectLimit = MAX_REDIRECT_LIMIT
    pagination = False
    pageFrom = 1
    offsetFrom = 1
    pageUnit = 0
    pageLimit = 0
    interval = None
    fromNow = None
    timeout = None
    ssl = None
    numRetries = 0
    delay = 1.
    interruptType = tuple()
    killType = tuple()
    errorType = tuple()
    responseType = "records"
    returnType = "records"
    mappedReturn = False
    root = list()
    groupby = list()
    groupSize = dict()
    countby = str()
    driver = None
    enableWire = False
    info = Info()
    flow = Flow()

    def __init__(self, fields: Optional[IndexLabel]=None, ranges: Optional[RangeFilter]=None, returnType: Optional[TypeHint]=None,
                tzinfo: Optional[Timezone]=None, countryCode=str(), datetimeUnit: Optional[Literal["second","minute","hour","day"]]=None,
                logName: Optional[str]=None, logLevel: LogLevel="WARN", logFile: Optional[str]=None, localSave=False,
                debugPoint: Optional[Keyword]=None, killPoint: Optional[Keyword]=None, extraSave: Optional[Keyword]=None,
                numRetries: Optional[int]=None, delay: Range=1., cookies: Optional[str]=None, numTasks=100,
                fromNow: Optional[Unit]=None, discard=True, progress=True, where=str(), which=str(), by=str(), message=str(),
                iterateUnit: Optional[int]=None, interval: Optional[Timedelta]=None, apiRedirect=False, redirectUnit: Optional[int]=None,
                queryList: GoogleQueryList=list(), uploadList: GoogleUploadList=list(), alertInfo: AlertInfo=dict(),
                driverOptions: Optional[Dict]=None, reopenDelay: Range=30., **context):
        AsyncSpider.__init__(self, **self.from_locals(locals(), drop=DRIVER_UNIQUE))
        self.set_driver_options(driverOptions, reopenDelay)

    def init_driver(func):
        @functools.wraps(func)
        async def wrapper(self: SeleniumSpider, *args, self_var=True, **context):
            args, context = self.init_context(args, context, self_var=self_var)
            self.run_driver(**self.get_driver_options(**context))
            try: data = await func(self, *args, **SESSION_CONTEXT(**context))
            finally: self.quit_driver()
            self.with_data(data, func=func.__name__, **context)
            return data
        return wrapper

    def with_driver(func):
        @functools.wraps(func)
        async def wrapper(self: SeleniumSpider, *args, **context):
            self.run_driver(**self.get_driver_options(**context))
            try: data = await func(self, *args, **context)
            finally: self.quit_driver()
            return data
        return wrapper

    async def interrupt(self, *args, exception: Exception, func: Callable, retryCount=0, reopenDelay: Range=30., **context) -> Tuple[int,Any]:
        if self.driver is not None:
            self.reopen_driver(self.get_driver_options(**context), reopenDelay)
            return await super().interrupt(*args, exception=exception, func=func, retryCount=retryCount, **context)
        else: raise exception

    def is_kill(self, exception: Exception) -> bool:
        return isinstance(exception, SELENIUM_SPIDER_KILL) or isinstance(exception, self.killType)


###################################################################
#################### Selenium Encrypted Spider ####################
###################################################################

class SeleniumEncryptedSpider(SeleniumSpider, EncryptedSpider):
    __metaclass__ = abc.ABCMeta
    asyncio = False
    operation = "spider"
    host = str()
    where = WHERE
    which = WHICH
    verb = ACTION
    by = str()
    fields = list()
    ranges = list()
    iterateArgs = list()
    iterateCount = dict()
    iterateProduct = list()
    iterateUnit = 1
    redirectUnit = 1
    pagination = False
    pageFrom = 1
    offsetFrom = 1
    pageUnit = 0
    pageLimit = 0
    interval = None
    fromNow = None
    timeout = None
    ssl = None
    numRetries = 0
    delay = 1.
    interruptType = tuple()
    killType = tuple()
    errorType = tuple()
    responseType = "records"
    returnType = "records"
    mappedReturn = False
    root = list()
    groupby = list()
    groupSize = dict()
    countby = str()
    auth = LoginSpider
    authKey = list()
    decryptedKey = dict()
    driver = None
    enableWire = False
    info = Info()
    flow = Flow()

    def __init__(self, fields: Optional[IndexLabel]=None, ranges: Optional[RangeFilter]=None, returnType: Optional[TypeHint]=None,
                tzinfo: Optional[Timezone]=None, countryCode=str(), datetimeUnit: Optional[Literal["second","minute","hour","day"]]=None,
                logName: Optional[str]=None, logLevel: LogLevel="WARN", logFile: Optional[str]=None, localSave=False,
                debugPoint: Optional[Keyword]=None, killPoint: Optional[Keyword]=None, extraSave: Optional[Keyword]=None,
                numRetries: Optional[int]=None, delay: Range=1., cookies: Optional[str]=None,
                fromNow: Optional[Unit]=None, discard=True, progress=True, where=str(), which=str(), by=str(), message=str(),
                iterateUnit: Optional[int]=None, interval: Optional[Timedelta]=None, apiRedirect=False, redirectUnit: Optional[int]=None,
                queryList: GoogleQueryList=list(), uploadList: GoogleUploadList=list(), alertInfo: AlertInfo=dict(),
                encryptedKey: Optional[EncryptedKey]=None, decryptedKey: Optional[DecryptedKey]=None,
                driverOptions: Optional[Dict]=None, reopenDelay: Range=30., **context):
        EncryptedSpider.__init__(self, **self.from_locals(locals(), drop=DRIVER_UNIQUE))
        self.set_driver_options(driverOptions, reopenDelay)

    def login_driver(func):
        @functools.wraps(func)
        def wrapper(self: SeleniumEncryptedSpider, *args, self_var=True, **context):
            args, context = self.init_context(args, context, self_var=self_var)
            options = self.get_driver_options(**context)
            with SeleniumWireDriver(**options) if self.enableWire else SeleniumDriver(**options) as driver:
                self.login(driver, **self.get_auth_key(update=True, **context))
                data = func(self, *args, driver=driver, **SESSION_CONTEXT(**self.set_auth_info(driver, **context)))
            time.sleep(.25)
            self.with_data(data, func=func.__name__, **context)
            return data
        return wrapper

    def is_kill(self, exception: Exception) -> bool:
        return isinstance(exception, SELENIUM_ENCRYPTED_SPIDER_KILL) or isinstance(exception, self.killType)


###################################################################
################# Selenium Encrypted Async Spider #################
###################################################################

class SeleniumEncryptedAsyncSpider(SeleniumAsyncSpider, EncryptedAsyncSpider):
    __metaclass__ = abc.ABCMeta
    asyncio = True
    operation = "spider"
    host = str()
    where = WHERE
    which = WHICH
    verb = ACTION
    by = str()
    fields = list()
    ranges = list()
    iterateArgs = list()
    iterateCount = dict()
    iterateProduct = list()
    iterateUnit = 1
    maxLimit = MAX_ASYNC_TASK_LIMIT
    redirectUnit = 1
    redirectLimit = MAX_REDIRECT_LIMIT
    pagination = False
    pageFrom = 1
    offsetFrom = 1
    pageUnit = 0
    pageLimit = 0
    interval = None
    fromNow = None
    timeout = None
    ssl = None
    numRetries = 0
    delay = 1.
    interruptType = tuple()
    killType = tuple()
    errorType = tuple()
    responseType = "records"
    returnType = "records"
    mappedReturn = False
    root = list()
    groupby = list()
    groupSize = dict()
    countby = str()
    auth = LoginSpider
    authKey = list()
    decryptedKey = dict()
    driver = None
    enableWire = False
    info = Info()
    flow = Flow()

    def __init__(self, fields: Optional[IndexLabel]=None, ranges: Optional[RangeFilter]=None, returnType: Optional[TypeHint]=None,
                tzinfo: Optional[Timezone]=None, countryCode=str(), datetimeUnit: Optional[Literal["second","minute","hour","day"]]=None,
                logName: Optional[str]=None, logLevel: LogLevel="WARN", logFile: Optional[str]=None, localSave=False,
                debugPoint: Optional[Keyword]=None, killPoint: Optional[Keyword]=None, extraSave: Optional[Keyword]=None,
                numRetries: Optional[int]=None, delay: Range=1., cookies: Optional[str]=None,
                fromNow: Optional[Unit]=None, discard=True, progress=True, where=str(), which=str(), by=str(), message=str(),
                iterateUnit: Optional[int]=None, interval: Optional[Timedelta]=None, apiRedirect=False, redirectUnit: Optional[int]=None,
                queryList: GoogleQueryList=list(), uploadList: GoogleUploadList=list(), alertInfo: AlertInfo=dict(),
                encryptedKey: Optional[EncryptedKey]=None, decryptedKey: Optional[DecryptedKey]=None,
                driverOptions: Optional[Dict]=None, reopenDelay: Range=30., **context):
        EncryptedAsyncSpider.__init__(self, **self.from_locals(locals(), drop=DRIVER_UNIQUE))
        self.set_driver_options(driverOptions, reopenDelay)

    def login_driver(func):
        @functools.wraps(func)
        async def wrapper(self: SeleniumEncryptedSpider, *args, self_var=True, **context):
            args, context = self.init_context(args, context, self_var=self_var)
            options = self.get_driver_options(**context)
            with SeleniumWireDriver(**options) if self.enableWire else SeleniumDriver(**options) as driver:
                self.login(driver, **self.get_auth_key(update=True, **context))
                data = await func(self, *args, driver=driver, **SESSION_CONTEXT(**self.set_auth_info(driver, **context)))
            await asyncio.sleep(.25)
            self.with_data(data, func=func.__name__, **context)
            return data
        return wrapper

    def is_kill(self, exception: Exception) -> bool:
        return isinstance(exception, SELENIUM_ENCRYPTED_SPIDER_KILL) or isinstance(exception, self.killType)



from setuptools import setup, find_packages
from gsdriver import __version__

REQUIRES = [
    "tqdm>=4.64.0",
    "asyncio>=3.4.3",
    "python-dateutil>=2.8.2",
    "pandas>=1.4.2",
    "openpyxl>=3.1.2",
    "bs4>=0.0.1",
    "requests>=2.31.1",
    "aiohttp>=3.9.1",
    "lxml>=4.9.4",
    "PyJWT>=2.9.0",
    "selenium>=4.8.2",
    "selenium-wire>=5.1.0",
    "blinker==1.7.0",
    "webdriver_manager>=4.0.1",
    "pyperclip>=1.8.2",
    "gspread>=5.6.2",
    "google-cloud-bigquery>=3.4.0",
    "holidays>=0.43",
    "workalendar>=17.0.0",
]

setup(
    name="gcp-webdriver",
    version=__version__,
    description="Selenium WebDriver utils with GCP functions",
    url="https://github.com/minyeamer/gsdriver.git",
    author="minyeamer",
    author_email="minyeamer@gmail.com",
    license="minyeamer",
    install_requires=REQUIRES,
    packages=find_packages(),
    keywords=["gcp-webdriver", "webdriver", "gcp"],
    python_requires=">=3.7",
)